import os
from pathlib import Path

import numpy as np

from rock_physics_open.equinor_utilities.snapshot_test_utilities import (
    INITIATE,
    compare_snapshots,
    get_snapshot_name,
    read_snapshot,
    store_snapshot,
)
from rock_physics_open.t_matrix_models import (
    run_t_matrix_forward_model_with_opt_params_exp,
    run_t_matrix_forward_model_with_opt_params_petec,
    run_t_matrix_with_opt_params_exp,
    run_t_matrix_with_opt_params_petec,
)

k_min = np.ones(101) * 71.0e9
mu_min = np.ones(101) * 32.0e9
rho_min = np.ones(101) * 2710.0
k_fl_orig = np.ones(101) * 2.7e9
rho_fl_orig = np.ones(101) * 1005.0
k_fl_sub = np.ones(101) * 0.8e9
rho_fl_sub = np.ones(101) * 800.0
vp = 4500.0 * np.linspace(1.1, 0.9, 101)
vs = 3000.0 * np.linspace(1.1, 0.9, 101)
rho = 2500 * np.linspace(1.1, 0.9, 101)
phi = np.linspace(0.15, 0.25, 101)
vsh = 0.1 * np.ones(101)
angle = 45.0
perm = 100.0
visco = 100.0
tau = 1e-7
freq = 1000
pressure = np.array([20.0e6, 22.0e6])


def test_run_t_matrix_with_opt_params_petec(data_dir: Path):
    fname = data_dir.joinpath("petec_opt_param.pkl")
    args = run_t_matrix_with_opt_params_petec(
        min_k=k_min,
        min_mu=mu_min,
        min_rho=rho_min,
        fl_k_orig=k_fl_orig,
        fl_rho_orig=rho_fl_orig,
        fl_k_sub=k_fl_sub,
        fl_rho_sub=rho_fl_sub,
        vp=vp,
        vs=vs,
        rhob=rho,
        phi=phi,
        angle=angle,
        perm=perm,
        visco=visco,
        tau=tau,
        freq=freq,
        f_name=fname,
        fluid_sub=True,
    )

    if not os.path.isfile(get_snapshot_name()) or INITIATE:
        _ = store_snapshot(get_snapshot_name(), *args)
    else:
        assert compare_snapshots(args, read_snapshot(get_snapshot_name()))


def test_run_t_matrix_with_opt_params_exp(data_dir: Path):
    fname = data_dir.joinpath("exp_opt_param.pkl")
    args = run_t_matrix_with_opt_params_exp(
        fl_k_orig=k_fl_orig,
        fl_rho_orig=rho_fl_orig,
        fl_k_sub=k_fl_sub,
        fl_rho_sub=rho_fl_sub,
        vp=vp,
        vs=vs,
        rhob=rho,
        phi=phi,
        vsh=vsh,
        angle=angle,
        perm=perm,
        visco=visco,
        tau=tau,
        freq=freq,
        f_name=fname,
        fluid_sub=True,
    )

    if not os.path.isfile(get_snapshot_name()) or INITIATE:
        _ = store_snapshot(get_snapshot_name(), *args)
    else:
        assert compare_snapshots(args, read_snapshot(get_snapshot_name()))


def test_run_t_matrix_opt_forward_model_petec(data_dir: Path):
    fname = data_dir.joinpath("petec_opt_param.pkl")
    args = run_t_matrix_forward_model_with_opt_params_petec(
        min_k=k_min,
        min_mu=mu_min,
        min_rho=rho_min,
        fl_k=k_fl_orig,
        fl_rho=rho_fl_orig,
        phi=phi,
        angle=angle,
        perm=perm,
        visco=visco,
        tau=tau,
        freq=freq,
        f_name=fname,
    )

    if not os.path.isfile(get_snapshot_name()) or INITIATE:
        _ = store_snapshot(get_snapshot_name(), *args)
    else:
        assert compare_snapshots(args, read_snapshot(get_snapshot_name()))


def test_run_t_matrix_opt_forward_model_exp(data_dir: Path):
    fname = data_dir.joinpath("exp_opt_param.pkl")
    args = run_t_matrix_forward_model_with_opt_params_exp(
        fl_k=k_fl_orig,
        fl_rho=rho_fl_orig,
        phi=phi,
        vsh=vsh,
        angle=angle,
        perm=perm,
        visco=visco,
        tau=tau,
        freq=freq,
        f_name=fname,
    )

    if not os.path.isfile(get_snapshot_name()) or INITIATE:
        _ = store_snapshot(get_snapshot_name(), *args)
    else:
        assert compare_snapshots(args, read_snapshot(get_snapshot_name()))



from importlib.resources import as_file, files
from typing import Any, Literal, overload

import numpy as np
import numpy.typing as npt
import scipy.optimize
from scipy.interpolate import RegularGridInterpolator

from rock_physics_open.equinor_utilities.conversions import celsius_to_kelvin
from rock_physics_open.equinor_utilities.various_utilities.types import Array2D

from .coefficients import (
    a0,
    theta0,
)
from .equations import residual_helmholtz_energy
from .tables.lookup_table import (
    load_lookup_table_interpolator,
)

# Constants
CO2_GAS_CONSTANT = 0.1889241 * 1e3  # J / kg K
CO2_CRITICAL_TEMPERATURE = 304.1282  # K
CO2_CRITICAL_DENSITY = 467.6  # kg / m^3
CO2_CRITICAL_PRESSURE = 7.3773  # MPa
CO2_TRIPLE_TEMPERATURE = 216.592  # K
CO2_TRIPLE_PRESSURE = 0.51795  # MPa


def co2_properties(
    temp: npt.NDArray[np.float64],
    pres: npt.NDArray[np.float64],
) -> tuple[npt.NDArray[np.float64], npt.NDArray[np.float64], npt.NDArray[np.float64]]:
    """
    CO2 properties are estimated according to Span & Wagner's equation of state model.
    References
    ----------
    R. Span and W. Wagner: A new equation of state for carbon dioxide covering the
    fluid region from the triple point temperature to 1100K at pressures up to
    800 MPa.
    J. Phys. Chem. Ref. Data, Vol. 25, No. 6, 1996, pp 1509 - 1596

    Parameters
    ----------
    temp: np.ndarray
        Temperature [°C]
    pres: np.ndarray
        Pressure [Pa]

    Returns
    -------
    tuple
        vel_co2, den_co2, k_co2 : (np.ndarray, np.ndarray, np.ndarray).
        vel_co2: velocity [m/s], den_co2: density [kg/m^3], k_co2: bulk modulus [Pa].
    """
    abs_temp = celsius_to_kelvin(temp)
    pres_mpa = pres * 1.0e-6
    den_co2 = carbon_dioxide_density(abs_temp, pres_mpa)
    k_co2 = carbon_dioxide_bulk_modulus(abs_temp, den_co2)
    vel_co2 = (k_co2 / den_co2) ** 0.5

    return vel_co2, den_co2, k_co2


def co2_helmholtz_energy(
    delta: npt.NDArray[np.float64],
    tau: npt.NDArray[np.float64],
    dd: Literal[0, 1, 2],
    dt: Literal[0, 1, 2],
) -> npt.NDArray[np.float64] | float:
    """
    Helmholtz energy as defined by equation 6.1 in Span & Wagner [2]

    :param delta: Reduced density, unit-less. That is, density / CO2_CRITICAL_DENSITY.
    :param tau: Inverse reduced temperature, unit-less. That is, CO2_CRITICAL_TEMPERATURE / (absolute) temperature
    :param dd: Degree of derivation wrt. delta. Integer between 0 and 2.
    :param dt: Degree of derivation wrt. tau. Integer between 0 and 2, as long as (dt + dd < 3)
    """
    return ideal_gas_helmholtz_energy(
        delta=delta,
        tau=tau,
        dd=dd,
        dt=dt,
    ) + co2_residual_helmholtz_energy(
        delta=delta,
        tau=tau,
        dd=dd,
        dt=dt,
    )


def ideal_gas_helmholtz_energy(
    delta: npt.NDArray[np.float64],
    tau: npt.NDArray[np.float64],
    dd: Literal[0, 1, 2],
    dt: Literal[0, 1, 2],
) -> npt.NDArray[np.float64] | float:
    """
    Helmholtz energy from ideal gas behavior as defined by equation 2.3 in Span & Wagner [2]. See function
    co2_helmholtz_energy for argument documentation.
    """
    # Adjust array shapes
    tau = np.asarray(tau)
    delta = np.asarray(delta)
    return_scalar = tau.ndim == 0
    tau2 = tau.reshape(-1, 1)  # Needed for array-based sums

    if dt == dd == 0:
        _sum = np.sum(a0[0, 3:] * np.log(1 - np.exp(-theta0 * tau2)), axis=-1)
        result = (
            np.log(delta) + a0[0, 0] + a0[0, 1] * tau + a0[0, 2] * np.log(tau) + _sum
        )
    elif dt == 1 and dd == 0:
        _sum = np.sum(
            a0[0, 3:] * theta0 * ((1 - np.exp(-theta0 * tau2)) ** -1 - 1), axis=-1
        )
        result = a0[0, 1] + a0[0, 2] / tau + _sum
    elif dt == 0 and dd == 1:
        return 1 / delta
    elif dt == 2 and dd == 0:
        _sum = np.sum(
            a0[0, 3:]
            * theta0**2
            * np.exp(-theta0 * tau2)
            * (1 - np.exp(-theta0 * tau2)) ** -2,
            axis=-1,
        )
        result = -a0[0, 2] / tau**2 - _sum
    elif dt == 0 and dd == 2:
        return -1 / delta**2
    elif dt == 1 and dd == 1:
        return 0
    else:
        raise ValueError
    if return_scalar:
        return result[0]
    return result


def co2_residual_helmholtz_energy(
    delta: npt.NDArray[np.float64],
    tau: npt.NDArray[np.float64],
    dd: Literal[0, 1, 2],
    dt: Literal[0, 1, 2],
) -> npt.NDArray[np.float64] | float:
    """
    Residual part of Helmholtz energy as defined by the equation in Table 32 of Span & Wagner [2]. See
    co2_helmholtz_energy for argument documentation.
    """
    tau = np.asarray(tau)
    delta = np.asarray(delta)
    return_scalar = (tau.ndim == 0) & (delta.ndim == 0)
    tau = tau.reshape(-1, 1)
    delta = delta.reshape(-1, 1)
    # tau == 1.0 or delta == 1.0 leads to numerically invalid results. The values are nudged to avoid nan output.
    tau[tau == 1.0] -= 1e-15
    delta[delta == 1.0] -= 1e-15

    res = residual_helmholtz_energy(
        delta_=delta,
        tau_=tau,
        diff_delta=dd,
        diff_tau=dt,
    )

    if return_scalar:
        return res[0]
    return res


@overload
def carbon_dioxide_pressure(
    absolute_temperature: npt.NDArray[np.float64],
    density: npt.NDArray[np.float64],
    d_density: Literal[0, 1] = ...,
    d_temperature: Literal[0, 1, 2] = ...,
    isentropic: bool = ...,
) -> npt.NDArray[np.float64]: ...


@overload
def carbon_dioxide_pressure(
    absolute_temperature: npt.NDArray[np.float64],
    density: npt.NDArray[np.float64],
    d_density: Literal[2],
    d_temperature: Literal[0, 1, 2] = ...,
    isentropic: bool = ...,
) -> None: ...


def carbon_dioxide_pressure(
    absolute_temperature: npt.NDArray[np.float64],
    density: npt.NDArray[np.float64],
    d_density: Literal[0, 1, 2] = 0,
    d_temperature: Literal[0, 1, 2] = 0,
    isentropic: bool = False,
) -> npt.NDArray[np.float64] | None:
    """
    CO2 pressure (MPa) as given by Table 3 of Span & Wagner [2]

    :param absolute_temperature: Temperature in K.
    :param density: CO2 density (kg / m^3).
    :param d_density: Degree of derivation wrt. density.
    :param d_temperature: Degree of derivation wrt. temperature.
    :param isentropic: Correction for isentropic conditions. Relevant primarily for isentropic bulk modulus
    """
    tau = CO2_CRITICAL_TEMPERATURE / absolute_temperature
    delta = density / CO2_CRITICAL_DENSITY
    if d_temperature != 0:
        raise ValueError(f"d_temperature must be 0, but was {d_temperature}")
    if d_density == 0:
        return (
            density
            * CO2_GAS_CONSTANT
            * absolute_temperature
            * (
                1
                + delta
                * co2_residual_helmholtz_energy(
                    delta=delta,
                    tau=tau,
                    dd=1,
                    dt=0,
                )
            )
            / 1e6
        )
    if d_density == 1:
        first = (
            2
            * delta
            * co2_residual_helmholtz_energy(
                delta=delta,
                tau=tau,
                dd=1,
                dt=0,
            )
        )
        second = delta**2 * co2_residual_helmholtz_energy(
            delta=delta,
            tau=tau,
            dd=2,
            dt=0,
        )
        if isentropic is False:
            third = 0
        else:
            # See Table 3 of Span & Wagner (speed of sound)
            nom = (
                1
                + delta
                * co2_residual_helmholtz_energy(
                    delta=delta,
                    tau=tau,
                    dd=1,
                    dt=0,
                )
                - delta
                * tau
                * co2_residual_helmholtz_energy(
                    delta=delta,
                    tau=tau,
                    dd=1,
                    dt=1,
                )
            ) ** 2
            den = tau**2 * (
                co2_helmholtz_energy(
                    delta=delta,
                    tau=tau,
                    dd=0,
                    dt=2,
                )
            )
            third = -nom / den
        return (
            absolute_temperature * CO2_GAS_CONSTANT * (1 + first + second + third) / 1e6
        )
    return None


def saturated_liquid_density(
    absolute_temperature: npt.NDArray[np.float64],
) -> npt.NDArray[np.float64]:
    """
    Saturated liquid density as defined by equation 3.14 of Span & Wagner [2]

    :param absolute_temperature: Absolute temperature in K. Should satisfy:
        CO2_TRIPLE_TEMPERATURE < absolute_temperature < CO2_CRITICAL_TEMPERATURE
    """
    _a1 = 1.9245108
    _a2 = -0.62385555
    _a3 = -0.32731127
    _a4 = 0.39245142
    _t = 1 - absolute_temperature / CO2_CRITICAL_TEMPERATURE
    inner = _a1 * _t**0.34 + _a2 * _t**0.5 + _a3 * _t ** (10 / 6) + _a4 * _t ** (11 / 6)
    return CO2_CRITICAL_DENSITY * np.exp(inner)


def saturated_vapor_density(
    absolute_temperature: npt.NDArray[np.float64],
) -> npt.NDArray[np.float64]:
    """
    Saturated vapor density as defined by equation 3.15 of Span & Wagner

    :param absolute_temperature: Absolute temperature in K. Should satisfy:
        CO2_TRIPLE_TEMPERATURE < absolute_temperature < CO2_CRITICAL_TEMPERATURE
    """
    # Assert temp < critical
    _a1 = -1.7074879
    _a2 = -0.82274670
    _a3 = -4.6008549
    _a4 = -10.111178
    _a5 = -29.742252
    _t = 1 - absolute_temperature / CO2_CRITICAL_TEMPERATURE
    inner = (
        _a1 * _t**0.34
        + _a2 * _t**0.5
        + _a3 * _t
        + _a4 * _t ** (7 / 3)
        + _a5 * _t ** (14 / 3)
    )
    return CO2_CRITICAL_DENSITY * np.exp(inner)


def sublimation_pressure(
    absolute_temperature: npt.NDArray[np.float64],
) -> npt.NDArray[np.float64]:
    """
    Sublimation pressure as defined by equation 3.12 of Span & Wagner [2]

    :param absolute_temperature: Absolute temperature in K. Should satisfy absolute_temperature < CO2_TRIPLE_TEMPERATURE
    """
    _a1 = -14.740846
    _a2 = 2.4327015
    _a3 = -5.3061778
    _t = 1 - absolute_temperature / CO2_TRIPLE_TEMPERATURE
    inner = _a1 * _t + _a2 * _t**1.9 + _a3 * _t**2.9
    return CO2_TRIPLE_PRESSURE * np.exp(inner / (1 - _t))


def vapor_pressure(
    absolute_temperature: npt.NDArray[np.float64],
) -> npt.NDArray[np.float64]:
    """
    Vapor pressure as defined by equation 3.13 of Span & Wagner [2]

    :param absolute_temperature: Absolute temperature in K. Should satisfy:
        CO2_TRIPLE_TEMPERATURE < absolute_temperature < CO2_CRITICAL_TEMPERATURE
    """
    _a1 = -7.0602087
    _a2 = 1.9391218
    _a3 = -1.6463597
    _a4 = -3.2995634
    _t = 1 - absolute_temperature / CO2_CRITICAL_TEMPERATURE
    inner = _a1 * _t**1.0 + _a2 * _t**1.5 + _a3 * _t**2.0 + _a4 * _t**4.0
    return CO2_CRITICAL_PRESSURE * np.exp(inner / (1 - _t))


def melting_pressure(
    absolute_temperature: npt.NDArray[np.float64],
) -> npt.NDArray[np.float64]:
    """
    Melting pressure as defined by equation 3.10 of Span & Wagner [2]

    :param absolute_temperature: Absolute temperature in K. Should satisfy CO2_TRIPLE_TEMPERATURE < absolute_temperature
    """
    _a1 = 1955.5390
    _a2 = 2055.4593
    _t = absolute_temperature / CO2_TRIPLE_TEMPERATURE - 1
    return CO2_TRIPLE_PRESSURE * (1 + _a1 * _t + _a2 * _t**2)


def _determine_density_bounds(
    absolute_temperature: npt.NDArray[np.float64],
    pressure: npt.NDArray[np.float64],
    force_vapor: bool | Literal["auto"],
) -> Array2D[np.float64]:
    """
    Calculate the upper and lower bound on density
    """
    bounds = np.zeros((absolute_temperature.size, 2))
    bounds[:, 0] = 0.1
    bounds[:, 1] = 1500.0

    below_triple = absolute_temperature < CO2_TRIPLE_TEMPERATURE
    below_critical = ~below_triple & (absolute_temperature < CO2_CRITICAL_TEMPERATURE)

    bounds[below_triple, 1] = saturated_vapor_density(
        absolute_temperature[below_triple]
    )
    if force_vapor is True:
        bounds[below_critical, 1] = saturated_vapor_density(
            absolute_temperature[below_critical]
        )
    elif force_vapor is False:
        bounds[below_critical, 0] = saturated_liquid_density(
            absolute_temperature[below_critical]
        )
    else:  # force_vapor == 'auto'
        below_vapor_pressure = pressure < vapor_pressure(absolute_temperature)
        is_vapor = below_critical & below_vapor_pressure
        bounds[is_vapor, 1] = saturated_vapor_density(absolute_temperature[is_vapor])
        is_liquid = below_critical & ~below_vapor_pressure
        bounds[is_liquid, 0] = saturated_liquid_density(absolute_temperature[is_liquid])
    return bounds


def _find_initial_density_values(
    bounds: Array2D[np.float64],
    absolute_temperature: npt.NDArray[np.float64],
    pressure: npt.NDArray[np.float64],
) -> npt.NDArray[np.float64]:
    """
    Finds approximate density values for the provided temperature(s) and pressure(s). The result is only intended to be
    used by array_carbon_dioxide_density.
    """

    temps = np.geomspace(
        np.min(absolute_temperature) * 0.99, np.max(absolute_temperature) * 1.01, 41
    )
    press = np.geomspace(
        np.min(pressure) * 0.99, np.max(absolute_temperature) * 1.01, 41
    )
    tt, pp = np.meshgrid(temps, press, indexing="ij")
    densi = carbon_dioxide_density(
        absolute_temperature=tt.flatten(),
        pressure=pp.flatten(),
        force_vapor="auto",
        raise_error=False,
    ).reshape(temps.size, press.size)
    points: Array2D[np.float64] = np.array((temps, press))
    rgi = RegularGridInterpolator(points, densi, method="linear")
    iv = rgi(np.array((absolute_temperature, pressure)).T)
    oob = (iv < bounds[:, 0]) | (iv > bounds[:, 1]) | np.isnan(iv)
    iv[oob] = np.mean(bounds[oob], axis=1)
    return iv


def array_carbon_dioxide_density(
    absolute_temperature: npt.NDArray[np.float64],
    pressure: npt.NDArray[np.float64],
    force_vapor: bool,
) -> npt.NDArray[np.float64]:
    """
    Alternative implementation of a vectorized carbon dioxide density function. Implemented primarily for demonstration
    purposes. For large arrays, a look-up-table approach should be preferred.

    Utilizes scipy.optimize.newton, which is the only root-finding method of scipy that supports a vectorized functions.

    For argument documentation, see carbon_dioxide_density.
    """
    absolute_temperature = np.asarray(absolute_temperature)
    pressure = np.asarray(pressure)
    bounds = _determine_density_bounds(absolute_temperature, pressure, force_vapor)
    iv = _find_initial_density_values(bounds, absolute_temperature, pressure)
    opt = scipy.optimize.newton(  # pyright: ignore[reportCallIssue, reportUnknownVariableType]
        lambda x: carbon_dioxide_pressure(absolute_temperature, x) - pressure,  # pyright: ignore[reportArgumentType]
        x0=iv,
        maxiter=10,
        full_output=True,
    )
    # scipy.optimize.newton may not always converge. We need to determine which of the elements of the solution are
    #  invalid. The opt.converged variable does not seem to suffice, so we perform separate checks. First, check that
    #  the solution is a valid root
    invalid = ~np.isclose(
        carbon_dioxide_pressure(
            absolute_temperature=absolute_temperature,
            density=opt.root,  # pyright: ignore[reportUnknownArgumentType, reportUnknownMemberType]
        ),
        pressure,
        atol=1e-5,
        rtol=0.0,
    )

    # Next, check if the solution is anywhere out of bounds (since newton does not support brackets), and check for nan
    # values.
    invalid |= (  # pyright: ignore[reportUnknownVariableType]
        (opt.root < bounds[:, 0]) | (opt.root > bounds[:, 1]) | np.isnan(opt.root)  # pyright: ignore[reportUnknownMemberType, reportUnknownArgumentType]
    )

    # Finally, use the robust density method to determine the invalid results
    sol = opt.root  # pyright: ignore[reportUnknownMemberType, reportUnknownVariableType]
    sol[invalid] = carbon_dioxide_density(
        absolute_temperature[invalid], pressure[invalid], force_vapor=force_vapor
    )
    return sol  # pyright: ignore[reportUnknownVariableType]


def _calculate_carbon_dioxide_density_scalar(
    absolute_temperature: float,
    pressure: float,
    force_vapor: bool | Literal["auto"] = "auto",
    raise_error: bool = True,
) -> float:
    """
    Density of carbon dioxide. Found solving the Pressure equation of Table 3 in Span & Wagner [2] numerically for
    density. To ensure a single solution, the phase of the liquid must first be determined.

    :param absolute_temperature: Absolute temperature (K).
    :param pressure: Pressure (MPa).
    :param force_vapor: Boolean or 'auto'. If 'auto', the phase of the fluid is automatically determined. However, along
        the vaporization line (assuming T_triple < absolute_temperature < T_critical), the fluid is in two-phase
        equilibrium and the phase cannot be uniquely determined. If force_vapor is set to True, vapor phase is always
        selected, if False, liquid phase is selected. Outside the temperature bounds, this argument has no effect. This
        argument should only be used close to the vaporization boundary, otherwise the behavior might not be as
        expected.
    :param raise_error: Boolean. If True, raises an error if density cannot be determined. Otherwise, returns np.nan.

    :return: Density (kg / m^3)
    """
    bounds = _determine_density_bounds(
        absolute_temperature=np.array([absolute_temperature]),
        pressure=np.array([pressure]),
        force_vapor=force_vapor,
    )[0, :]

    # Extend bounds slightly to ensure toms748 can converge
    bounds = (bounds[0] * 0.95, bounds[1] * 1.05)

    try:
        opt = scipy.optimize.root_scalar(  # pyright: ignore[reportCallIssue, reportUnknownVariableType]
            lambda x: carbon_dioxide_pressure(absolute_temperature, x) - pressure,  # pyright: ignore[reportArgumentType]
            method="toms748",
            bracket=bounds,
            x0=np.sum(bounds) / 2,
        )
    except ValueError:
        if raise_error:
            raise
        return np.nan
    return opt.root  # pyright: ignore[reportUnknownMemberType, reportUnknownVariableType]


_calculate_carbon_dioxide_density = np.vectorize(
    _calculate_carbon_dioxide_density_scalar, otypes=[float]
)


def carbon_dioxide_density(
    absolute_temperature: npt.NDArray[np.float64],
    pressure: npt.NDArray[np.float64],
    interpolate: bool = False,
    **kwargs: Any,
) -> npt.NDArray[np.float64]:
    """
    Density of carbon dioxide. Found either by direct calculation or interpolation. Any additional arguments are passed
    to _calculate_carbon_dioxide_density.

    :param absolute_temperature: Absolute temperature (K).
    :param pressure: Pressure (MPa).
    :param interpolate: Flag whether to interpolate data or not. If not, data is calculated directly. This is more
        accurate, but also more time-consuming. Data outside the bounds of the interpolator will be set to np.nan.

    :return: Density (kg / m^3)
    """
    if interpolate is False:
        return _calculate_carbon_dioxide_density(
            absolute_temperature=absolute_temperature,
            pressure=pressure,
            **kwargs,
        )
    assert interpolate is True

    ref = (
        files("rock_physics.fluid_models.gas_model.span_wagner.tables")
        / "carbon_dioxide_density.npz"
    )
    with as_file(ref) as fp:
        interpolator = load_lookup_table_interpolator(fp)

    return interpolator(absolute_temperature, pressure)


def carbon_dioxide_bulk_modulus(
    absolute_temperature: npt.NDArray[np.float64],
    density: npt.NDArray[np.float64],
) -> npt.NDArray[np.float64]:
    """
    Isentropic bulk modulus, derived from the expression for speed of sound in Table 3 of Span & Wagner
    """
    d_pressure = carbon_dioxide_pressure(
        absolute_temperature=absolute_temperature,
        density=density,
        d_density=1,
        isentropic=True,
    )
    return density * d_pressure * 1e6



import os

import numpy as np

from rock_physics_open.equinor_utilities.snapshot_test_utilities import (
    INITIATE,
    compare_snapshots,
    get_snapshot_name,
    read_snapshot,
    store_snapshot,
)
from rock_physics_open.sandstone_models import friable_model, friable_model_dry
from rock_physics_open.sandstone_models.friable_models import CoordinateNumberFunction

k_min = 36.8e9 * np.ones(20)
mu_min = 44.0e9 * np.ones(20)
rho_min = 2650 * np.ones(20)
k_fl = 2.7e9 * np.ones(20)
rho_fl = 1005 * np.ones(20)
phi = np.linspace(0.0, 0.50, 20)
p_eff = 20.0e6 * np.ones(20)
phi_c = 0.45
coord_num_func: CoordinateNumberFunction = "PorBased"
n = 8.0
shear_red = 0.25


def test_friable_model():
    args = friable_model(
        k_min=k_min,
        mu_min=mu_min,
        rho_min=rho_min,
        k_fl=k_fl,
        rho_fl=rho_fl,
        phi=phi,
        p_eff=p_eff,
        phi_c=phi_c,
        coord_num_func=coord_num_func,
        n=n,
        shear_red=shear_red,
    )

    if not os.path.isfile(get_snapshot_name()) or INITIATE:
        _ = store_snapshot(get_snapshot_name(), *args)
    else:
        assert compare_snapshots(args, read_snapshot(get_snapshot_name()))


def test_friable_model_dry():
    args = friable_model_dry(
        k_min=k_min,
        mu_min=mu_min,
        phi=phi,
        p_eff=p_eff,
        phi_c=phi_c,
        coord_num_func=coord_num_func,
        n=n,
        shear_red=shear_red,
    )

    if not os.path.isfile(get_snapshot_name()) or INITIATE:
        _ = store_snapshot(get_snapshot_name(), *args)
    else:
        assert compare_snapshots(args, read_snapshot(get_snapshot_name()))



import sys
from typing import Any, Literal, cast

import numpy as np

from rock_physics_open.equinor_utilities import gen_utilities
from rock_physics_open.equinor_utilities.various_utilities.types import (
    Array1D,
    Array2D,
    TMatrixCallable,
)

from .t_matrix_C import t_matrix_porosity_c_alpha_v


def parse_t_matrix_inputs(
    k_min: Array1D[np.float64],
    mu_min: Array1D[np.float64],
    rho_min: Array1D[np.float64],
    k_fl: Array1D[np.float64],
    rho_fl: Array1D[np.float64],
    phi: Array1D[np.float64],
    perm: Array1D[np.float64] | float,
    visco: Array1D[np.float64] | float,
    alpha: Array1D[np.float64] | Array2D[np.float64],
    v: Array1D[np.float64] | Array2D[np.float64],
    tau: Array1D[np.float64],
    frequency: float,
    angle: float,
    frac_inc_con: Array1D[np.float64] | float,
    frac_inc_ani: Array1D[np.float64] | float,
    pressure: Array1D[np.float64] | None,
    scenario: int | None,
    fcn: TMatrixCallable | str | None,
) -> tuple[
    Array1D[np.float64],
    Array1D[np.float64],
    Array1D[np.float64],
    Array1D[np.float64],
    Array1D[np.float64],
    Array1D[np.float64],
    Array1D[np.float64],
    Array1D[np.float64],
    Array1D[np.float64],
    Array2D[np.float64],
    Array2D[np.float64],
    float,
    float,
    Array1D[np.float64],
    Array1D[np.float64],
    Array1D[np.float64] | None,
    TMatrixCallable,
    Literal[0, 1, 2],
    Literal[0, 1, 2],
]:
    """Function to do all necessary checking of input type, dimension, reshaping etc. that clutter up the start of T-Matrix
    NB: Setting scenario will override the settings for alpha, v and tau.

    Parameters
    ----------
    k_min : np.ndarray
        N length numpy array, mineral bulk modulus [Pa].
    mu_min : np.ndarray
        N length numpy array, mineral shear modulus [Pa].
    rho_min : np.ndarray
        N length numpy array, mineral density [kg/m^3].
    k_fl : np.ndarray
        N length numpy array, fluid bulk modulus [Pa].
    rho_fl : np.ndarray
        N length numpy array, mineral density [kg/m^3].
    phi : np.ndarray or float
        N length numpy array, total porosity [ratio].
    perm : np.ndarray or float
        float or N length numpy array, permeability [mD].
    visco : np.ndarray or float
        float or N length numpy array, fluid viscosity [cP].
    alpha : np.ndarray
        M or NxM length numpy array, inclusion aspect ratio [ratio].
    v : np.ndarray
        M or NxM length numpy array, inclusion concentration [ratio].
    tau : np.ndarray
        M length numpy array, relaxation time [s].
    frequency : float
        single float, signal frequency [Hz].
    angle : float
        single float, angle of symmetry plane [degree].
    frac_inc_con : np.ndarray or float
        single float or N length numpy array, fraction of connected inclusions [ratio].
    frac_inc_ani : np.ndarray or float
        single float or N length numpy array, fraction of anisotropic inclusions [ratio].
    pressure : list or np.ndarray
         > 1 value list or numpy array in ascending order, effective pressure [Pa].
    scenario : int
        pre-set scenarios for alpha, v and tau
    fcn : callable | str | None.
        function with which to run the T-Matrix model or string with function name within t_matrix_models. If None, the C++ implementation is used.

    Returns
    -------
    tuple
        All inputs in correct dimension and data type plus ctrl_connected, ctrl_anisotropy - control parameters.
    """

    def _assert_type(
        arg: Any,
        exp_dtype: type | tuple[type, ...],
        err_str: str = "check inputs, wrong type encountered",
    ) -> None:
        """Assert type.

        Parameters
        ----------
        arg : any
            To be asserted.
        exp_dtype : type
            Expected data type.
        err_str : str, optional
            Error string, by default 'check inputs, wrong type encountered'.

        Raises
        ------
        ValueError
            Content of err_str.
        """
        if not isinstance(arg, exp_dtype):
            raise ValueError("t-matrix inputs: " + err_str)

    # 1: Check all single float values, silently cast int to float
    # Permeability and viscosity: Check that they are floats and convert them to to SI units
    _assert_type(
        arg=perm,
        exp_dtype=(float, int),
        err_str="expect permeability given as single float in units mD",
    )
    perm = perm * 0.986923e-15

    _assert_type(
        arg=visco,
        exp_dtype=(float, int),
        err_str="expect viscosity given as single float in units cP",
    )
    visco = visco * 1.0e-2

    _assert_type(
        arg=frequency,
        exp_dtype=(float, int),
        err_str="expect frequency given as single float value in Hz",
    )
    frequency = float(frequency)

    _assert_type(
        arg=angle,
        exp_dtype=(float, int),
        err_str="expect angle given as single float value in degrees",
    )
    angle = float(angle)

    # 2: Determine the T-Matrix function, use the C++ implementation as default in case of None
    # If it given as a string, it must belong to the t_matrix_models module
    if not fcn:  # None
        t_matrix_fcn = t_matrix_porosity_c_alpha_v
    elif not callable(fcn):
        fcn_err_str = (
            "T-Matrix function should be given as the callable function or a string "
            "to the function name within t_matrix_models "
        )
        _assert_type(fcn, str, fcn_err_str)
        # Import here to avoid circular import
        import importlib

        t_matrix_models = importlib.import_module("rock_physics_open.t_matrix_models")

        if not hasattr(t_matrix_models, fcn):
            raise ValueError(fcn_err_str)
        t_matrix_fcn = getattr(t_matrix_models, fcn)

    else:  # Function must be recognised
        if not fcn:  # check that the function exists
            raise ValueError(
                "t-matrix inputs: function for T-Matrix model is not known"
            )
        t_matrix_fcn = fcn

    # 3: Check all inputs that should be one value per sample, expand perm and visco to match this.
    # frac_inc_ani and frac_inc_con can be single floats or vectors of the same length as the other vector inputs

    log_length = len(phi)

    # Check that the inputs that should have the same length actually do
    # dim_check_vector may throw an error message, modify it to show origin
    try:
        (
            k_min,
            mu_min,
            rho_min,
            k_fl,
            rho_fl,
            phi,
            perm,
            visco,
            frac_inc_con,
            frac_inc_ani,
        ) = cast(  # Casting since dim_check_vector typing is incomplete
            list[Array1D[np.float64]],
            gen_utilities.dim_check_vector(
                (
                    k_min,
                    mu_min,
                    rho_min,
                    k_fl,
                    rho_fl,
                    phi,
                    perm,
                    visco,
                    frac_inc_con,
                    frac_inc_ani,
                ),
                force_type=np.dtype(float),
            ),
        )
    except ValueError:
        raise ValueError("t-matrix inputs: {}".format(str(sys.exc_info())))

    # 4: Scenario will override settings for alpha, v and tau
    if scenario:
        scenario = int(scenario)
        if scenario == 1:  # Mostly rounded pores
            alpha = np.array([0.9, 0.1])
            v = np.array([0.9, 0.1])
        elif scenario == 2:  # Dual porosity with little rounded pores
            alpha = np.array([0.58, 0.027])
            v = np.array([0.85, 0.15])
        elif scenario == 3:  # Mixed pores
            alpha = np.array([0.9, 0.1, 0.01])
            v = np.array([0.8, 0.19, 0.01])
        elif scenario == 4:  # Flat pores and cracks
            alpha = np.array([0.9, 0.1, 0.01, 0.001])
            v = np.array([0.689, 0.3, 0.01, 0.001])
        tau = np.ones_like(alpha) * 1.0e-7

    # 4a: Check alpha, v, tau
    # alpha and v is either 1D vector with inclusion aspect ratios and proportions applying to all samples or 2D arrays
    # where the first dimension must coincide with the log_length
    # tau should be a 1D vector with length equal to the number of inclusions per sample
    _assert_type(
        arg=alpha,
        exp_dtype=np.ndarray,
        err_str="alpha should be a 1D or 2D numpy array of matching size with v",
    )
    _assert_type(
        arg=v,
        exp_dtype=np.ndarray,
        err_str="v should be a 1D or 2D numpy array og matching size with alpha",
    )
    _assert_type(
        arg=tau,
        exp_dtype=(float, np.ndarray),
        err_str="tau should be a single float or 1D numpy array with length matching the number of inclusions",
    )

    alpha_shape = alpha.shape
    # First make sure that alpha and v have the same number of elements
    try:
        alpha, v = cast(  # casting since dim_check_vector typing is incomplete
            list[Array1D[np.float64]],
            gen_utilities.dim_check_vector((alpha, v), force_type=np.dtype(float)),
        )
    except ValueError:
        raise ValueError("t-matrix inputs: {}".format(str(sys.exc_info())))
    alpha = alpha.reshape(alpha_shape)  # pyright: ignore[reportAssignmentType] | Should be same shape as before
    v = v.reshape(alpha_shape)  # pyright: ignore[reportAssignmentType] | Should be same shape as before
    # If they are 2D arrays, the first dimension should match log_length
    if alpha.ndim == 2:
        if not (alpha_shape[0] == log_length or alpha_shape[0] == 1):
            raise ValueError(
                "t-matrix inputs: number of samples in alpha is not matching length of inputs"
            )
        if alpha_shape[0] == 1:
            alpha = np.tile(alpha, (log_length, 1))
            v = np.tile(v, (log_length, 1))

    elif alpha.ndim == 1:
        alpha = np.tile(alpha.reshape(1, alpha_shape[0]), (log_length, 1))
        v = np.tile(v.reshape(1, alpha_shape[0]), (log_length, 1))
    # Check that the number of elements in tau matches dim 1 of alpha
    if isinstance(tau, float):
        tau = np.ones(alpha_shape) * tau
    else:
        if not tau.ndim == 1 and tau.shape[0] == alpha.shape[1]:  # pyright: ignore[reportGeneralTypeIssues] | alpha is Array2D
            raise ValueError(
                "t-matrix inputs: number of elements in tau is not matching number of inclusions"
            )

    # 5: Check pressure
    # Check that pressure either consists of at least two elements in ascending order or is None
    # in the case there is no pressure modelling
    if pressure is not None:
        _assert_type(
            arg=pressure,
            exp_dtype=(list, np.ndarray),
            err_str="pressure should be a list or numpy array of at least two elements in ascending order",
        )
        pressure = np.sort(np.array(pressure).flatten())
        if not pressure.shape[0] > 1:
            raise ValueError(
                "t-matrix inputs: if pressure is given, it should contain two or more values"
            )

    # 6: Determine case for connected/isolated and isotropic/anisotropic inclusions
    # Assume the most general case: mix of isotropic and anisotropic inclusions and connected and isolated inclusions
    ctrl_anisotropy = 1
    if np.all(frac_inc_ani == 0):
        ctrl_anisotropy = 0  # All isotropic
        angle = 0
    elif np.all(frac_inc_ani == 1):
        ctrl_anisotropy = 2  # All anisotropic

    ctrl_connected = 1
    # Create case based on amount of connected inclusions
    if np.all(frac_inc_con == 0):
        ctrl_connected = 0  # All isolated
    elif np.all(frac_inc_con == 1):
        ctrl_connected = 2  # All connected

    return (
        k_min,
        mu_min,
        rho_min,
        k_fl,
        rho_fl,
        phi,
        perm,
        visco,
        alpha,  # pyright: ignore[reportReturnType] | alpha is Array1D or Array2D
        v,
        tau,
        frequency,
        angle,
        frac_inc_con,
        frac_inc_ani,
        pressure,
        t_matrix_fcn,
        ctrl_connected,
        ctrl_anisotropy,
    )



# ruff: noqa: UP034
"""
Comprehensive pytest test suite for pressure sensitivity models.

This module tests all model types with proper fixtures, parameterization,
and mock testing for external rock physics dependencies.

pytest test_pressure_models.py -v                    # Run all tests
pytest test_pressure_models.py -m unit              # Run only unit tests
pytest test_pressure_models.py -m benchmark         # Run only benchmarks
pytest test_pressure_models.py -k "exponential"     # Run tests matching pattern

"""

import os
import tempfile
from typing import Any

import numpy as np
import numpy.typing as npt
import pytest

# Import all model classes
from rock_physics_open.equinor_utilities.machine_learning_utilities.exponential_model import (
    ExponentialPressureModel,
)
from rock_physics_open.equinor_utilities.machine_learning_utilities.polynomial_model import (
    PolynomialPressureModel,
)
from rock_physics_open.equinor_utilities.machine_learning_utilities.sigmoidal_model import (
    SigmoidalPressureModel,
)


# Fixtures
@pytest.fixture
def exponential_model() -> ExponentialPressureModel:
    """Create exponential model for testing."""
    return ExponentialPressureModel(
        a_factor=0.5,
        b_factor=1e7,
        model_max_pressure=5e7,
        description="Test exponential model",
    )


@pytest.fixture
def polynomial_model() -> PolynomialPressureModel:
    """Create polynomial model for testing."""
    return PolynomialPressureModel(
        weights=[1.0, 2e-8, -1e-16],
        model_max_pressure=5e7,
        description="Test polynomial model",
    )


@pytest.fixture
def sigmoidal_model() -> SigmoidalPressureModel:
    """Create sigmoidal model for testing."""
    return SigmoidalPressureModel(
        phi_amplitude=1000.0,
        phi_median_point=0.2,
        phi_x_scaling=10,
        phi_bias=2000.0,
        p_eff_median_point=1.5e7,
        p_eff_x_scaling=1e-7,
        p_eff_bias=1000.0,
        model_max_pressure=5e7,
        description="Test sigmoidal model",
    )


@pytest.fixture
def exp_poly_valid_input() -> npt.NDArray[np.float64]:
    """Valid input for exponential and polynomial models."""
    return np.array([[3000.0, 2e7, 1e7], [3200.0, 2.5e7, 1.2e7]])


@pytest.fixture
def sigmoidal_valid_input():
    """Valid input for sigmoidal model."""
    return np.array([[0.2, 2e7, 1e7], [0.25, 2.5e7, 1.2e7]])


# Invalid data fixtures for error testing
@pytest.fixture
def invalid_input_data() -> dict[str, Any]:
    """Provide various invalid input formats for testing."""
    rng = np.random.default_rng(42)
    return {
        "wrong_type": [[1, 2, 3], [4, 5, 6]],  # List instead of ndarray
        "wrong_dimensions": np.array([1, 2, 3, 4, 5]),  # 1D instead of 2D
        "wrong_columns_exp": rng.random((10, 4)),  # 4 columns instead of 3
        "empty_array": np.empty((0, 3)),  # Empty array
    }


# Test ExponentialPressureModel
class TestExponentialPressureModel:
    """Test cases for ExponentialPressureModel."""

    def test_initialization(self, exponential_model: ExponentialPressureModel):
        """Test model initialization."""
        assert exponential_model.a_factor == 0.5
        assert exponential_model.b_factor == 1e7
        assert exponential_model.max_pressure == 5e7
        assert exponential_model.description == "Test exponential model"

    def test_validate_input_valid(
        self,
        exponential_model: ExponentialPressureModel,
        exp_poly_valid_input: npt.NDArray[np.float64],
    ):
        """Test input validation with valid data."""
        result = exponential_model.validate_input(exp_poly_valid_input)
        np.testing.assert_array_equal(result, exp_poly_valid_input)

    def test_validate_input_invalid(
        self,
        exponential_model: ExponentialPressureModel,
        invalid_input_data: dict[str, Any],
    ):
        """Test validation with invalid input formats."""
        test_cases = [
            ("wrong_type", "Input must be numpy ndarray."),
            ("wrong_dimensions", "Input must be \\(n,3\\)"),
            ("wrong_columns_exp", "Input must be \\(n,3\\)"),
            ("empty_array", None),  # Empty array might be valid
        ]

        for case_name, expected_error in test_cases:
            if expected_error is None:
                continue  # Skip cases where we don't expect an error

            with pytest.raises(ValueError, match=expected_error):
                _ = exponential_model.validate_input(invalid_input_data[case_name])

    @pytest.mark.parametrize("case", ["in_situ", "depleted"])
    def test_predict_abs(
        self,
        exponential_model: ExponentialPressureModel,
        exp_poly_valid_input: npt.NDArray[np.float64],
        case: str,
    ):
        """Test absolute prediction for different cases."""
        result = exponential_model.predict_abs(exp_poly_valid_input, case=case)
        assert len(result) == 2
        assert np.all(result > 0)

    def test_predict_differential(
        self,
        exponential_model: ExponentialPressureModel,
        exp_poly_valid_input: npt.NDArray[np.float64],
    ):
        """Test differential prediction."""
        result = exponential_model.predict(exp_poly_valid_input)
        assert len(result) == 2

    def test_predict_max(
        self,
        exponential_model: ExponentialPressureModel,
        exp_poly_valid_input: npt.NDArray[np.float64],
    ):
        """Test prediction with max pressure."""
        result = exponential_model.predict_max(exp_poly_valid_input)
        assert len(result) == 2

    def test_predict_max_no_max_pressure(
        self, exp_poly_valid_input: npt.NDArray[np.float64]
    ):
        """Test predict_max without max_pressure set."""
        model_no_max = ExponentialPressureModel(0.5, 1e7)
        with pytest.raises(ValueError, match='Field "model_max_pressure" is not set'):
            _ = model_no_max.predict_max(exp_poly_valid_input)

    def test_todict(self, exponential_model: ExponentialPressureModel):
        """Test dictionary conversion."""
        result = exponential_model.todict()
        expected_keys = {"a_factor", "b_factor", "model_max_pressure", "description"}
        assert set(result.keys()) == expected_keys

    def test_save_load(self, exponential_model: ExponentialPressureModel):
        """Test save and load functionality."""
        with tempfile.NamedTemporaryFile(delete=False, suffix=".pkl") as tmp:
            tmp_path = tmp.name

        try:
            exponential_model.save(tmp_path)
            loaded_model = ExponentialPressureModel.load(tmp_path)

            assert loaded_model.a_factor == exponential_model.a_factor
            assert loaded_model.b_factor == exponential_model.b_factor
            assert loaded_model.max_pressure == exponential_model.max_pressure
        finally:
            os.unlink(tmp_path)


# Test PolynomialPressureModel
class TestPolynomialPressureModel:
    """Test cases for PolynomialPressureModel."""

    def test_initialization(self, polynomial_model: PolynomialPressureModel):
        """Test model initialization."""
        assert polynomial_model.weights == [1.0, 2e-8, -1e-16]

    def test_predict_abs_no_weights(
        self, exp_poly_valid_input: npt.NDArray[np.float64]
    ):
        """Test prediction with empty weights."""
        model_no_weights = PolynomialPressureModel([])
        with pytest.raises(ValueError, match='Field "weights" is not set'):
            _ = model_no_weights.predict_abs(exp_poly_valid_input)

    def test_predict_abs(
        self,
        polynomial_model: PolynomialPressureModel,
        exp_poly_valid_input: npt.NDArray[np.float64],
    ):
        """Test absolute prediction."""
        result = polynomial_model.predict_abs(exp_poly_valid_input, case="in_situ")
        assert len(result) == 2
        assert np.all(result > 0)


# Test SigmoidalPressureModel
class TestSigmoidalPressureModel:
    """Test cases for SigmoidalPressureModel."""

    def test_properties(self, sigmoidal_model: SigmoidalPressureModel):
        """Test model properties."""
        assert sigmoidal_model.phi_amplitude == 1000.0
        assert sigmoidal_model.phi_median_point == 0.2
        assert sigmoidal_model.p_eff_median_point == 1.5e7

    def test_predict_abs(
        self,
        sigmoidal_model: SigmoidalPressureModel,
        sigmoidal_valid_input: npt.NDArray[np.float64],
    ):
        """Test absolute prediction."""
        result = sigmoidal_model.predict_abs(sigmoidal_valid_input, case="in_situ")
        assert len(result) == 2
        assert np.all(result > 0)

    def test_input_validation(self, sigmoidal_model: SigmoidalPressureModel):
        """Test input validation with wrong format."""
        invalid_input = np.array([[0.2, 2e7]])  # Missing column
        with pytest.raises(ValueError, match="Input must be \\(n,3\\)"):
            _ = sigmoidal_model.validate_input(invalid_input)


# Integration Tests
class TestModelIntegration:
    """Integration tests across model types."""

    @pytest.mark.parametrize(
        (
            "model_fixture",
            "input_fixture",
        ),
        [
            ("exponential_model", "exp_poly_valid_input"),
            ("polynomial_model", "exp_poly_valid_input"),
            ("sigmoidal_model", "sigmoidal_valid_input"),
        ],
    )
    def test_all_models_predict_interface(
        self, request: pytest.FixtureRequest, model_fixture: str, input_fixture: str
    ):
        """Test that all models implement the required interface."""
        model = request.getfixturevalue(model_fixture)
        test_input = request.getfixturevalue(input_fixture)

        # Test interface methods exist and work
        assert hasattr(model, "predict_abs")
        assert hasattr(model, "predict")
        assert hasattr(model, "predict_max")
        assert hasattr(model, "validate_input")
        assert hasattr(model, "todict")
        assert hasattr(model, "save")

        result_abs = model.predict_abs(test_input)
        result_diff = model.predict(test_input)
        result_max = model.predict_max(test_input)

        assert len(result_abs) == len(test_input)
        assert len(result_diff) == len(test_input)
        assert len(result_max) == len(test_input)

    def test_model_serialization_consistency(
        self,
        exponential_model: ExponentialPressureModel,
        polynomial_model: PolynomialPressureModel,
        sigmoidal_model: SigmoidalPressureModel,
    ):
        """Test that all models can be serialized and deserialized consistently."""
        models = [exponential_model, polynomial_model, sigmoidal_model]

        for model in models:
            # Test dictionary conversion
            model_dict = model.todict()
            assert isinstance(model_dict, dict)
            assert len(model_dict) > 0

            # Test save/load cycle
            with tempfile.NamedTemporaryFile(delete=False, suffix=".pkl") as tmp:
                tmp_path = tmp.name

            try:
                model.save(tmp_path)
                loaded_model = type(model).load(tmp_path)

                # Compare dictionaries
                original_dict = model.todict()
                loaded_dict = loaded_model.todict()

                assert original_dict == loaded_dict

            finally:
                os.unlink(tmp_path)


# Performance Tests
class TestModelPerformance:
    """Performance and stress tests."""

    @pytest.mark.parametrize("n_samples", [100, 1000, 10000])
    def test_exponential_model_performance(
        self, exponential_model: ExponentialPressureModel, n_samples: int
    ):
        """Test exponential model performance with different data sizes."""
        # Generate test data
        rng = np.random.default_rng(42)
        velocities = rng.uniform(2500, 4000, n_samples)
        p_in_situ = rng.uniform(1e7, 3e7, n_samples)
        p_depleted = rng.uniform(0.5e7, 1.5e7, n_samples)
        test_data = np.column_stack([velocities, p_in_situ, p_depleted])

        # Test prediction (should complete without errors)
        result = exponential_model.predict_abs(test_data)
        assert len(result) == n_samples
        assert np.all(np.isfinite(result))

    def test_memory_usage(self, exponential_model: ExponentialPressureModel):
        """Test that models don't leak memory with repeated calls."""
        # Create moderate size dataset
        n_samples = 10000
        rng = np.random.default_rng(42)
        velocities = rng.uniform(2500, 4000, n_samples)
        p_in_situ = rng.uniform(1e7, 3e7, n_samples)
        p_depleted = rng.uniform(0.5e7, 1.5e7, n_samples)
        test_data = np.column_stack([velocities, p_in_situ, p_depleted])

        # Run multiple predictions
        for _ in range(10):
            result = exponential_model.predict_abs(test_data)
            del result  # Explicitly delete to test memory cleanup



from .display_result_statistics import disp_result_stats
from .gassmann_dry_mod import gassmann_dry_model
from .gassmann_mod import gassmann_model
from .gassmann_sub_mod import gassmann_sub_model
from .hs_average import hs_average
from .pressure import pressure
from .reflectivity import reflectivity
from .timeshift import time_shift_pp, time_shift_ps
from .vp_vs_rho_set_statistics import vp_vs_rho_stats
from .vrh_3_min import min_3_voigt_reuss_hill

__all__ = [
    "disp_result_stats",
    "gassmann_dry_model",
    "gassmann_model",
    "gassmann_sub_model",
    "hs_average",
    "pressure",
    "reflectivity",
    "time_shift_pp",
    "time_shift_ps",
    "vp_vs_rho_stats",
    "min_3_voigt_reuss_hill",
]



import numpy as np
import numpy.typing as npt

NULL_CLASS = 0


def norm_class(
    obs: npt.NDArray[np.float64],
    class_mean: npt.NDArray[np.float64],
    class_cov: npt.NDArray[np.float64],
    prior_prob: npt.NDArray[np.float64],
    class_id: npt.NDArray[np.int64],
    thresh: float = np.inf,
) -> tuple[npt.NDArray[np.float64], npt.NDArray[np.float64], npt.NDArray[np.float64]]:
    """
    Normal distribution classification routine. All data points are assigned a
    class, unless a threshold is set. The "dist" calculated here is the quadratic
    discriminant function according to a Bayes classification. This is a negative
    number, and the closest class has the smallest absolute value.

    Parameters
    ----------
    obs : np.ndarray
        An nxm array, where n is the number of samples and m is the number of variables.
    class_mean : np.ndarray
        A pxm array, where p is the number of classes and m is the number of variables.
    class_cov : np.ndarray
        A pxm array, where p is the number of classes and m is the number of variables.
    prior_prob : np.ndarray
        A p length vector, where p is the number of classes containing prior probabilities for each class.
    class_id : np.ndarray
        A p length vector, where p is the number of classes, containing class_id (integer numbers).
    thresh : float
        Unclassified threshold.


    Returns
    --------
    tuple
        norm_class_id, norm_dist, norm_pp : (np.ndarray, np.ndarray, np.ndarray).
        norm_class_id:	nx1 vector. The classes are numbered 1 to m, and unclassified
        samples (with absolute distance greater than thresh) are set to 0,
        norm_dist: nx1 vector with quadratic discriminant distance from the closest class centre to sample,
        norm_pp: nx1 vector with posterior probability based on the distance to each class.
    """

    # Find dimensions
    n = obs.shape[0]
    p = class_mean.shape[0]

    # Assign matrices
    dist = np.zeros((n, p))

    # Calculate distance for each class
    for i in range(p):
        cov_inv = np.linalg.inv(class_cov[:, :, i])
        delta = obs - class_mean[i, :]
        dist[:, i] = (
            -0.5 * np.log(np.linalg.det(class_cov[:, :, i]))
            - 0.5 * np.sqrt(np.einsum("nj,jk,nk->n", delta, cov_inv, delta))
            + np.log(prior_prob[i])
        )

    # The discrimination function ("dist") are negative numbers. Choose the one
    # with the smallest value as the closest class
    norm_class_id = np.choose(np.argmax(dist, axis=1), class_id)
    norm_dist = np.amax(dist, axis=1)
    norm_pp = -np.exp(norm_dist) / np.sum(np.exp(dist), axis=1)

    # Compare the absolute value of the discriminator with the threshold
    norm_class_id[np.abs(norm_dist) > thresh] = NULL_CLASS

    return norm_class_id, norm_dist, norm_pp



from typing import Any, overload

import numpy as np
import numpy.typing as npt
import pandas as pd


@overload
def dim_check_vector(
    args: list[Any] | tuple[Any, ...],
    force_type: np.dtype | None = ...,
) -> list[npt.NDArray[Any] | pd.DataFrame]:
    """Overload for when the input is a list or tuple."""


@overload
def dim_check_vector(
    args: pd.DataFrame,
    force_type: np.dtype | None = ...,
) -> pd.DataFrame:
    """Overload for when the input is a pandas DataFrame."""


@overload
def dim_check_vector(
    args: npt.NDArray[Any],
    force_type: np.dtype | None = ...,
) -> npt.NDArray[Any]:
    """Overload for when the input is a numpy array."""


def dim_check_vector(
    args: list[Any] | tuple[Any, ...] | npt.NDArray[Any] | pd.DataFrame,
    force_type: np.dtype | None = None,
) -> npt.NDArray[Any] | pd.DataFrame | list[npt.NDArray[Any] | pd.DataFrame]:
    """
    Check that all inputs are of the same (one-dimensional) size. Raise ValueError in case there are several lengths
    present in the inputs. All inputs will be checked and possibly expanded to common length. Only the first dimension
    is harmonised.

    Parameters
    ----------
    args : list or tuple
        Input list or tuple of scalars, numpy arrays or pandas data frames of numerical or boolean type.
    force_type : np.dtype
        Force all outputs to be of a specific dtype.

    Returns
    -------
    output_args : list
        List of inputs where all are of the same length.
    """
    single_types = (np.ndarray, pd.DataFrame)
    iterable_types = (list, tuple)
    allowed_types = single_types + iterable_types
    if not isinstance(args, allowed_types):  # pyright: ignore[reportUnnecessaryIsInstance] | Kept for backward compatibility
        raise ValueError("dim_check_vector: unknown input type: {}".format(type(args)))  # pyright: ignore[reportUnreachable] | Kept for backward compatibility

    # Single array or dataframe is just returned
    if isinstance(args, single_types):
        if force_type is not None:
            try:
                args = args.astype(force_type)
            except ValueError:
                raise ValueError(
                    "dim_check_vector: not possible to force dtype to {}".format(
                        force_type
                    )
                )
        return args

    # If any input is a scalar, make it into an array
    if force_type is not None:
        try:
            args = [
                np.array(item, ndmin=1, dtype=force_type)
                if np.isscalar(item)
                else item.astype(force_type)
                for item in args
            ]
        except ValueError:
            raise ValueError(
                "dim_check_vector: not possible to force dtype to {}".format(force_type)
            )
    else:
        args = [np.array(item, ndmin=1) if np.isscalar(item) else item for item in args]

    # Can now test for length - must either be a scalar or have the same length
    max_length: int = np.max([item.shape[0] for item in args])
    if not np.all([item.shape[0] == max_length or item.shape[0] == 1 for item in args]):
        raise ValueError(
            "dim_check_vector: Unequal array lengths in input to dim_check_vector"
        )

    output_arg: list[npt.NDArray[Any] | pd.DataFrame] = []
    for item in args:
        if item.shape[0] == max_length:
            output_arg.append(item)
        else:
            item_dim = item.ndim
            repeat_tuple = tuple([max_length] + [1] * (item_dim - 1))
            if isinstance(item, pd.DataFrame):
                output_arg.append(
                    pd.DataFrame(
                        np.tile(np.array(item), repeat_tuple),
                        columns=item.columns,
                        index=np.arange(max_length),
                    )
                )
            else:
                output_arg.append(np.tile(item, repeat_tuple))

    return output_arg



from .brine_model import brine_properties
from .gas_model import gas_properties
from .oil_model import oil_properties

__all__ = [
    "brine_properties",
    "gas_properties",
    "oil_properties",
]



from typing import Literal

import numpy as np

from rock_physics_open.equinor_utilities.various_utilities.types import (
    Array1D,
    Array2D,
    Array3D,
)

from .array_functions import array_inverse, array_matrix_mult
from .g_tensor import g_tensor_vec
from .iso_av import iso_av_vec


def calc_isolated_part_vec(
    c0: Array3D[np.float64],
    s_0: Array3D[np.float64],
    kappa_f: Array1D[np.float64],
    alpha: Array2D[np.float64],
    v: Array2D[np.float64],
    case_iso: Literal[0, 1, 2],
    frac_ani: float,
) -> Array3D[np.float64]:
    """
    Returns the first order correction tensor: sum of the concentrations and
    t-matrices of the isolated porosity (6x6 matrix).
    case_iso = 0 : 100% isotropic,
    case_iso = 1 : mixed isotropic and anisotropic porosity,
    case_iso = 2 : 100% anisotropic porosity.

    Parameters
    ----------
    c0 : np.ndarray
        Stiffness tensor of the host material (6x6xn).
    s_0 :  np.ndarray
        Inverse of stiffness tensor.
    kappa_f : np.ndarray
        Bulk modulus of the fluid (n length vector).
    alpha : np.ndarray
        Aspect ratios of all the inclusions (1xnumber of inclusions) vector).
    v : np.ndarray
        Concentration of all the inclusions (1xnumber of inclusions) vector).
    case_iso : int
        Control parameter.
    frac_ani : float
        Fraction of anisotropic inclusions.

    Returns
    -------
    np.ndarray
        c1: correction tensor.

    Notes
    -----
    29.10.2020 HFLE: Simplification: alpha for isotropic and anisotropic part are the same. This has been default in
    Remy's code, but this function had assumption that half of the inclusions could have different aspect ratio
    13.11.2020 HFLE: In case of zero porosity, this routine returns a zero tensor, whereas the correct should have
    been to return the host material tensor - corrected.

    """
    if not (c0.ndim == 3 and s_0.ndim == 3):
        raise ValueError(f"{__name__}: mismatch in inputs variables dimension/shape")

    log_length = c0.shape[0]
    if v.ndim != 2:
        v = np.tile(v.reshape(1, v.shape[0]), (log_length, 1))

    if alpha.ndim == 1 and alpha.shape[0] != c0.shape[0]:
        alpha = np.tile(alpha.reshape(1, alpha.shape[0]), (log_length, 1))
    alpha_len = alpha.shape[1]

    cn = np.zeros_like(c0)
    cn[:, 0:3, 0:3] = np.tile(kappa_f.reshape((log_length, 1, 1)), (1, 3, 3))
    cn_d = cn - c0
    i4 = np.tile(np.eye(6).reshape(1, 6, 6), (log_length, 1, 1))
    c1 = np.zeros_like(c0)

    # Will need G tensor for each alpha
    g_arr = [g_tensor_vec(c0=c0, s_0=s_0, alpha=alpha[:, i]) for i in range(alpha_len)]

    if case_iso != 1:
        # if there is only isotropic or anisotropic inclusions
        for j in range(alpha_len):
            t = array_matrix_mult(
                cn_d, array_inverse(i4 - array_matrix_mult(g_arr[j], cn_d))
            )
            if case_iso != 2:
                t = iso_av_vec(t)
            c1 = c1 + v[:, j].reshape(log_length, 1, 1) * t
    else:
        # Isotropic and anisotropic part
        for j in range(alpha_len):
            t = array_matrix_mult(
                cn_d, array_inverse(i4 - array_matrix_mult(g_arr[j], cn_d))
            )
            t = iso_av_vec(t)
            c1 = c1 + ((1 - frac_ani) * v[:, j]).reshape(log_length, 1, 1) * t
        for j in range(alpha_len):
            t = array_matrix_mult(
                cn_d, array_inverse(i4 - array_matrix_mult(g_arr[j], cn_d))
            )
            c1 = c1 + (frac_ani * v[:, j]).reshape(log_length, 1, 1) * t

    idx_zero = np.sum(v, axis=1) == 0.0
    if np.any(idx_zero):
        c1[idx_zero] = c0[idx_zero]

    return c1



from sys import byteorder
from typing import Any

import numpy as np
import numpy.typing as npt
import pandas as pd

WRONG_BYTEORDER = ">" if byteorder == "little" else "<"


def filter_input_log(
    args: list[Any] | tuple[Any, ...] | npt.NDArray[Any] | pd.DataFrame,
    working_int: npt.NDArray[Any] | None = None,
    negative: bool = False,
    no_zero: bool = False,
    positive: bool = True,
) -> tuple[npt.NDArray[np.bool_], list[npt.NDArray[Any] | pd.DataFrame]]:
    """
    Check for valid input values in numpy arrays or pandas data frames. Default behaviour is to
    identify missing values - assumed to be NaN and Inf. Other conditions
    can be stated in the key word arguments. Unknown conditions are ignored and a warning
    is issued. Run dim_check_vector to make sure that all inputs have the same length.
    Erroneous values in a sample in one log will remove the sample from all the logs.
    All inputs must have the same array length (data frames the same number of indices).

    Parameters
    ----------
    args : list or tuple or np.ndarray or pd.DataFrame
        Inputs to be filtered, single array or dataframe or lists of arrays or data frames.
    working_int : np.ndarray
        Valid positions are shown as values > 0.
    negative : bool
        Positive values are excluded (zero values are retained).
    no_zero : bool
        Zero values are excluded.
    positive : bool
        Negative values are excluded.

    Returns
    -------
    tuple
        idx, output_args : (np.ndarray, list)
        indices of valid values [bool],
        list of input arrays at valid indices.
    """
    type_error = "filter_input_log: unknown input data type: {}".format(type(args))
    size_error = "filter_input_log: inputs of different length"

    if not isinstance(args, (list, tuple, np.ndarray, pd.DataFrame)):  # pyright: ignore[reportUnnecessaryIsInstance] | Kept for backward compatibility
        raise ValueError(type_error)  # pyright: ignore[reportUnreachable] | Kept for backward compatibility
    # Make sure that 'args' is iterable
    if isinstance(args, (np.ndarray, pd.DataFrame)):
        args = [args]

    # Input tuple
    if isinstance(args, tuple):
        args = list(args)

    # Need to preserve original inputs
    input_args = args.copy()

    # Test that inputs are of the right types and the same length
    if not np.all([isinstance(log, (np.ndarray, pd.DataFrame)) for log in args]):
        raise ValueError(type_error)
    if not np.all([log.shape[0] == args[0].shape[0] for log in args]):
        raise ValueError(size_error)

    # Generate pandas series from numpy arrays
    args_: list[pd.Series | pd.DataFrame] = [
        pd.Series(log) if isinstance(log, np.ndarray) else log for log in args
    ]
    # Merge into a data frame
    logs = pd.concat(args_, axis=1)

    # If any of the input logs are of type boolean, False means that they should not be included,
    # regardless of filter flags
    bool_col = logs.dtypes == "bool"
    if any(bool_col):
        idx = ~logs.any(bool_only=True, axis=1)
        logs.drop(columns=logs.columns[bool_col], inplace=True)
    else:
        idx = pd.Series(index=logs.index, data=np.zeros_like(logs.index).astype(bool))

    # Standard checks: NaN and Inf
    idx = np.logical_or(idx, logs.isna().any(axis=1))
    idx = np.logical_or(idx, logs.isin([np.inf, -np.inf]).any(axis=1))

    # Remove columns with dtype that is not numeric
    obj_col = [dd.kind not in ["i", "u", "f", "c"] for dd in logs.dtypes]
    logs.drop(columns=logs.columns[obj_col], inplace=True)

    # Checks according to the input options input_dict
    # Only consider working interval if it is included or set to some value
    if working_int is not None and not np.all(working_int == 0):
        idx = np.logical_or(idx, working_int == 0)
    if negative:
        # noinspection PyTypeChecker
        idx = np.logical_or(idx, (logs >= 0.0).all(axis=1))
        # idx = np.logical_or(idx, logs.loc[logs > 0.0]).any(axis=1)
    if no_zero:
        idx = np.logical_or(idx, (logs == 0.0).any(axis=1))
    if positive:
        # noinspection PyTypeChecker
        idx = np.logical_or(idx, (logs < 0.0).any(axis=1))

    # Negate idx to identify samples to retain
    idx = np.logical_not(idx)
    num_valid_samples = idx.sum()
    if num_valid_samples == 0:
        raise ValueError("No acceptable input values")
    for i in range(len(input_args)):
        if isinstance(input_args[i], np.ndarray):
            input_args[i] = input_args[i][idx]
        else:  # data frame
            # https://pandas.pydata.org/pandas-docs/stable/user_guide/gotchas.html#byte-ordering-issues

            check_type = (
                np.array([col_type.byteorder for col_type in input_args[i].dtypes])
                == WRONG_BYTEORDER
            )
            if np.any(check_type):
                tmp_array = (
                    input_args[i].to_numpy().byteswap().newbyteorder().astype(float)
                )
                cols = input_args[i].columns
                for j in range(check_type.shape[0]):
                    if check_type[j]:
                        input_args[i][cols[j]] = tmp_array[:, j]
            input_args[i] = input_args[i].loc[idx]
    return np.array(idx), input_args



import numpy as np
from numpy.random import default_rng

from rock_physics_open.equinor_utilities.classification_functions import poly_class


class TestPolyClass:
    def test_poly_class(self):
        rg = default_rng(234769238476)
        obs = rg.random((11, 2))
        class_poly = np.array(
            [
                [
                    [0.0, 0.0],
                    [0.5, 0.0],
                    [0.5, 0.5],
                    [0.0, 0.5],
                    [0.0, 0.0],
                ],
                [
                    [0.5, 0.0],
                    [1.0, 0.0],
                    [1.0, 0.5],
                    [0.5, 0.5],
                    [0.5, 0.0],
                ],
                [
                    [0.0, 0.5],
                    [1.0, 0.5],
                    [1.0, 1.0],
                    [0.0, 0.5],
                    [0.0, 0.0],
                ],
            ]
        )
        class_id = np.array([1, 2, 3])
        poly_class_arr = poly_class(obs, class_poly, class_id)
        poly_class_ref = np.array([1, 3, 1, 2, 2, 1, 3, 2, 0, 2, 1])
        np.testing.assert_equal(poly_class_arr, poly_class_ref)



import numpy as np
import numpy.typing as npt

from rock_physics_open.equinor_utilities import std_functions


def gassmann_sub_model(
    k_min: npt.NDArray[np.float64],
    k_fl_orig: npt.NDArray[np.float64],
    rho_fl_orig: npt.NDArray[np.float64],
    k_fl_sub: npt.NDArray[np.float64],
    rho_fl_sub: npt.NDArray[np.float64],
    k_sat_orig: npt.NDArray[np.float64],
    mu: npt.NDArray[np.float64],
    rho_sat_orig: npt.NDArray[np.float64],
    por: npt.NDArray[np.float64],
) -> tuple[
    npt.NDArray[np.float64],
    npt.NDArray[np.float64],
    npt.NDArray[np.float64],
    npt.NDArray[np.float64],
    npt.NDArray[np.float64],
    npt.NDArray[np.float64],
    npt.NDArray[np.float64],
]:
    """
    Gassmann model to go from one saturated state to another.

    Parameters
    ----------
    k_min : np.ndarray
        Mineral bulk modulus [Pa].
    k_fl_orig : np.ndarray
        Original fluid  bulk modulus [Pa].
    rho_fl_orig : np.ndarray
        Original fluid density [lg/m^3].
    k_fl_sub : np.ndarray
        Substituted fluid  bulk modulus [Pa].
    rho_fl_sub : np.ndarray
        Substituted fluid density [lg/m^3].
    k_sat_orig : np.ndarray
        Saturated rock bulk modulus with original fluid  [Pa].
    mu : np.ndarray
        Rock shear modulus [Pa].
    rho_sat_orig : np.ndarray
        Saturated rock density with original fluid [kg/m^3].
    por : np.ndarray
        Porosity [fraction].

    Returns
    -------
    tuple
        vp_sat, vs_sat, rho_sat, ai_sat, vpvs_sat, k_sat, mu : np.ndarray
        vp_sat, vs_sat:  saturated velocities [m/s], rho_sat: saturated density [kg/m^3], ai_sat: saturated acoustic
        impedance [kg/m^3 x m/s], vpvs_sat: saturated velocity ratio [unitless], k_sat, mu: saturated bulk modulus and
        shear modulus (the latter unchanged from dry state) [Pa].
    """
    rho_sat_sub = rho_sat_orig + por * (rho_fl_sub - rho_fl_orig)
    k_sat_sub = std_functions.gassmann2(k_sat_orig, k_fl_orig, k_fl_sub, por, k_min)
    vp_sat_sub, vs_sat_sub, ai_sat_sub, vpvs_sat_sub = std_functions.velocity(
        k_sat_sub, mu, rho_sat_sub
    )

    return vp_sat_sub, vs_sat_sub, rho_sat_sub, ai_sat_sub, vpvs_sat_sub, k_sat_sub, mu



import numpy as np
import numpy.typing as npt

NULL_CLASS = 0


def lin_class(
    obs: npt.NDArray[np.float64],
    class_mean: npt.NDArray[np.float64],
    class_id: npt.NDArray[np.float64],
    thresh: float = np.inf,
) -> tuple[npt.NDArray[np.float64], npt.NDArray[np.float64]]:
    """
    Linear classification routine. All data points are assigned a class, unless a threshold is set.

    Parameters
    ----------
    obs : np.ndarray
        An nxm array, where n is the number of samples and m is the number of features.
    class_mean : np.ndarray
        A pxm array, where p is the number of classes and m is the number of features.
    class_id : np.ndarray
        A p length vector, where p is the number of classes, containing class_id (integer numbers).
    thresh : float
        Unclassified threshold.

    Returns
    -------
    tuple
        lin_class_arr, lin_dist : (np.ndarray, np.ndarray).
        lin_class_arr: nx1 vector. The classes are numbered according to class_id,
        and unclassified samples (with distance greater than thresh) are set to 0,
        lin_dist: nx1 vector with linear distance from the closest class centre to each sample.
    """

    # Find dimensions
    n = obs.shape[0]
    p = class_mean.shape[0]

    # Assign matrices
    dist = np.zeros((n, p))

    # Calculate distance for each class
    for i in range(p):
        dist[:, i] = np.sqrt(np.sum((obs - class_mean[i, :]) ** 2, axis=1))

    # Find the shortest distance, assign class, filter out observations with distance
    # greater than the threshold
    lin_class_arr = np.choose(np.argmin(dist, axis=1), class_id)
    lin_dist = np.amin(dist, axis=1)
    lin_class_arr[lin_dist > thresh] = NULL_CLASS

    return lin_class_arr, lin_dist



import numpy as np
import numpy.typing as npt

from rock_physics_open.equinor_utilities import std_functions


def gassmann_model(
    k_min: npt.NDArray[np.float64],
    k_fl: npt.NDArray[np.float64],
    rho_fl: npt.NDArray[np.float64],
    k_dry: npt.NDArray[np.float64],
    mu: npt.NDArray[np.float64],
    rho_dry: npt.NDArray[np.float64],
    por: npt.NDArray[np.float64],
) -> tuple[
    npt.NDArray[np.float64],
    npt.NDArray[np.float64],
    npt.NDArray[np.float64],
    npt.NDArray[np.float64],
    npt.NDArray[np.float64],
    npt.NDArray[np.float64],
    npt.NDArray[np.float64],
]:
    """
    Gassmann model to go from dry rock to saturated state.

    Parameters
    ----------
    k_min : np.ndarray
        Mineral bulk modulus [Pa].
    k_fl : np.ndarray
        Fluid  bulk modulus [Pa].
    rho_fl : np.ndarray
        Fluid density [lg/m^3].
    k_dry : np.ndarray
        Dry rock bulk modulus [Pa].
    mu : np.ndarray
        Dry rock shear modulus [Pa].
    rho_dry : np.ndarray
        Dry rock density [kg/m^3].
    por : np.ndarray
        Porosity [fraction].

    Returns
    -------
    tuple
        vp_sat, vs_sat, rho_sat, ai_sat, vpvs_sat, k_sat, mu : np.ndarray
        vp_sat, vs_sat:  saturated velocities [m/s], rho_sat: saturated density [kg/m^3], ai_sat: saturated acoustic
        impedance [kg/m^3 x m/s], vpvs_sat: saturated velocity ratio [unitless], k_sat, mu: saturated bulk modulus and
        shear modulus (the latter unchanged from dry state) [Pa].
    """
    rho_sat = rho_dry + por * rho_fl
    k_sat = std_functions.gassmann(k_dry, por, k_fl, k_min)
    vp_sat, vs_sat, ai_sat, vpvs_sat = std_functions.velocity(k_sat, mu, rho_sat)

    return vp_sat, vs_sat, rho_sat, ai_sat, vpvs_sat, k_sat, mu



import numpy as np
import numpy.typing as npt


def dvorkin_contact_cement(
    frac_cem: npt.NDArray[np.float64] | float,
    por0_sst: npt.NDArray[np.float64] | float,
    mu0_sst: npt.NDArray[np.float64],
    k0_sst: npt.NDArray[np.float64],
    mu0_cem: npt.NDArray[np.float64],
    k0_cem: npt.NDArray[np.float64],
    vs_red: npt.NDArray[np.float64] | float,
    c: float,
) -> tuple[npt.NDArray[np.float64], npt.NDArray[np.float64]]:
    """
    Dvorkin-Nur contact cement model for estimation of elastic moduli.

    Parameters
    ----------
    frac_cem : np.ndarray | float
        Cement fraction of volume [ratio].
    por0_sst : np.ndarray | float
        Critical porosity of sand [ratio].
    mu0_sst : np.ndarray
        Mineral shear modulus of sand [Pa].
    k0_sst : np.ndarray
        Mineral bulk modulus of sand [Pa].
    mu0_cem : np.ndarray
        Mineral shear modulus of cement [Pa].
    k0_cem : np.ndarray
        Mineral bulk modulus of cement [Pa].
    vs_red : np.ndarray | float
        Shear modulus reduction factor [ratio].
    c : np.ndarray | float
        Coordination number (grain contacts per grain) [unitless].

    Returns
    -------
    tuple
        k_cc, mu_cc : numpy.ndarray.
        k_cc: bulk modulus [Pa], mu_cc: shear modulus [Pa].
    """
    alpha = (2 * frac_cem / (3 * (1 - por0_sst))) ** 0.5
    poiss = (3 * k0_sst - 2 * mu0_sst) / (2 * (3 * k0_sst + mu0_sst))
    poiss_c = (3 * k0_cem - 2 * mu0_cem) / (2 * (3 * k0_cem + mu0_cem))
    a_an = (2 * mu0_cem / (np.pi * mu0_sst)) * (
        (1 - poiss) * (1 - poiss_c) / (1 - 2 * poiss_c)
    )
    a_at = mu0_cem / (np.pi * mu0_sst)

    a_t = (
        -1e-2
        * (2.26 * poiss**2 + 2.07 * poiss + 2.3)
        * a_at ** (0.079 * poiss**2 + 0.1754 * poiss - 1.342)
    )
    b_t = (0.0573 * poiss**2 + 0.0937 * poiss + 0.202) * a_at ** (
        0.0274 * poiss**2 + 0.0529 * poiss - 0.8765
    )
    c_t = (
        1e-4
        * (9.654 * poiss**2 + 4.945 * poiss + 3.1)
        * a_at ** (0.01867 * poiss**2 + 0.4011 * poiss - 1.8186)
    )

    s_t = a_t * alpha**2 + b_t * alpha + c_t

    c_n = 0.00024649 * a_an ** (-1.9864)
    b_n = 0.20405 * a_an ** (-0.89008)
    a_n = -0.024153 * a_an ** (-1.3646)

    s_n = a_n * alpha**2 + b_n * alpha + c_n

    m0_cem = k0_cem + 4 / 3 * mu0_cem
    k_cc = (1 / 6) * c * (1 - por0_sst) * m0_cem * s_n
    mu_cc = (3 / 5) * k_cc + vs_red * (3 / 20) * c * (1 - por0_sst) * mu0_cem * s_t

    return k_cc, mu_cc



import numpy as np
import numpy.typing as npt

from rock_physics_open.equinor_utilities import std_functions


def gassmann_dry_model(
    k_min: npt.NDArray[np.float64],
    k_fl: npt.NDArray[np.float64],
    rho_fl: npt.NDArray[np.float64],
    k_sat: npt.NDArray[np.float64],
    mu: npt.NDArray[np.float64],
    rho_sat: npt.NDArray[np.float64],
    por: npt.NDArray[np.float64],
) -> tuple[
    npt.NDArray[np.float64],
    npt.NDArray[np.float64],
    npt.NDArray[np.float64],
    npt.NDArray[np.float64],
    npt.NDArray[np.float64],
    npt.NDArray[np.float64],
    npt.NDArray[np.float64],
]:
    """
    Gassmann model to go from saturated rock to dry state.

    Parameters
    ----------
    k_min : np.ndarray
        Mineral bulk modulus [Pa].
    k_fl : np.ndarray
        Fluid  bulk modulus [Pa].
    rho_fl : np.ndarray
        Fluid density [lg/m^3].
    k_sat : np.ndarray
        Saturated rock bulk modulus [Pa].
    mu : np.ndarray
        Saturated rock shear modulus [Pa].
    rho_sat : np.ndarray
        Saturated rock density [kg/m^3].
    por : np.ndarray
        Porosity [fraction].

    Returns
    -------
    tuple
        vp_dry, vs_dry, rho_dry, ai_dry, vpvs_dry, k_dry, mu : np.ndarray
        vp_dry, vs_dry:  dry velocities [m/s], rho_dry: dry density [kg/m^3], ai_dry: dry acoustic impedance
        [kg/m^3 x m/s], vpvs_dry: dry velocity ratio [unitless], k_dry, mu: dry bulk modulus and shear modulus (the
        latter unchanged from saturated state) [Pa].
    """
    rho_dry = rho_sat - por * rho_fl
    k_dry = std_functions.gassmann_dry(k_sat, por, k_fl, k_min)
    vp_dry, vs_dry, ai_dry, vpvs_dry = std_functions.velocity(k_dry, mu, rho_dry)

    return vp_dry, vs_dry, rho_dry, ai_dry, vpvs_dry, k_dry, mu



import numpy as np
from numpy.random import default_rng

from rock_physics_open.equinor_utilities.various_utilities import (
    time_shift_pp,
    time_shift_ps,
)


class TestTimeShift:
    def test_time_shift_ps(self):
        rg = default_rng(672190547)
        tvd = np.linspace(1900.0, 2100.0, 11)
        vp_base = 3500.0 * (0.5 + 0.5 * rg.random(11))
        vp_mon = 3500.0 * (0.5 + 0.5 * rg.random(11))
        vs_base = 1700.0 * (0.5 + 0.5 * rg.random(11))
        vs_mon = 1700.0 * (0.5 + 0.5 * rg.random(11))
        multiplier = 1
        twt_ps_shift = time_shift_ps(tvd, vp_base, vp_mon, vs_base, vs_mon, multiplier)
        twt_ref = np.array(
            [
                -1.487565,
                -1.8224189,
                1.0883962,
                0.0848656,
                2.6961336,
                -2.1949398,
                -4.3835154,
                0.2291774,
                -6.2099461,
                -5.7735105,
                -9.4191945,
            ]
        )
        np.testing.assert_almost_equal(twt_ps_shift, twt_ref)

    def test_time_shift_pp(self):
        rg = default_rng(3874629384)
        tvd = np.linspace(1900.0, 2100.0, 11)
        vp_base = 3500.0 * (0.5 + 0.5 * rg.random(11))
        vp_mon = 3500.0 * (0.5 + 0.5 * rg.random(11))
        multiplier = 1
        owt_pp_shift, twt_pp_shift = time_shift_pp(tvd, vp_base, vp_mon, multiplier)
        owt_ref = np.array(
            [
                -2.389087,
                -2.6406689,
                -1.1021602,
                -4.4961946,
                -0.858652,
                -2.7415121,
                -5.375477,
                -3.7926991,
                -6.0184832,
                -2.2550444,
                -5.1210616,
            ]
        )
        twt_ref = 2.0 * owt_ref
        np.testing.assert_almost_equal(owt_pp_shift, owt_ref)
        np.testing.assert_almost_equal(twt_pp_shift, twt_ref)



from typing import cast

import numpy as np
import numpy.typing as npt

from rock_physics_open.equinor_utilities import std_functions
from rock_physics_open.equinor_utilities.gen_utilities import dim_check_vector

from .dem import dem_model


def shale_4_min_dem_overlay(
    k1: npt.NDArray[np.float64],
    mu1: npt.NDArray[np.float64],
    rho1: npt.NDArray[np.float64],
    k2: npt.NDArray[np.float64],
    mu2: npt.NDArray[np.float64],
    rho2: npt.NDArray[np.float64],
    k3: npt.NDArray[np.float64],
    mu3: npt.NDArray[np.float64],
    rho3: npt.NDArray[np.float64],
    k4: npt.NDArray[np.float64],
    mu4: npt.NDArray[np.float64],
    rho4: npt.NDArray[np.float64],
    k_fl: npt.NDArray[np.float64],
    rho_fl: npt.NDArray[np.float64],
    phi: npt.NDArray[np.float64],
    f1: npt.NDArray[np.float64],
    f2: npt.NDArray[np.float64],
    prop_clay: float,
    asp: npt.NDArray[np.float64],
) -> tuple[npt.NDArray[np.float64], npt.NDArray[np.float64], npt.NDArray[np.float64]]:
    """
    Simple shale model with mixture of matrix minerals in Voigt-Reuss-Hill
    and DEM inclusion model for fluid filled porosity. The model is aimed at
    overlays in cross-plots where two input fractions are given, and the two
    remaining are given as a proportion.
    Fraction of carbonate and clay is implicit 1 - f1 - f2. f1 and f2 can be set to zero
    All k, mu values in [Pa], rho in [kg/m^3], f, phi, prop_clay, asp in [fraction].

    Parameters
    ----------
    k1, mu1, rho1 : np.ndarray
        Mineral 1 properties [Quartz and feldspar].
    k2, mu2, rho2 : np.ndarray
        Mineral 2 properties [Kerogen].
    k3, mu3, rho3 : np.ndarray
        Mineral 3 properties [Clay].
    k4, mu4, rho4 : np.ndarray
        Mineral 4 properties [Carbonates].
    k_fl, rho_fl : np.ndarray
        Fluid properties.
    phi : np.ndarray
        Porosity.
    f1 : np.ndarray
        Fraction of mineral 1.
    f2 : np.ndarray
        Fraction of mineral 2.
    prop_clay : float
        Range 0 - 1 of the fraction that is clay and carbonate.
    asp: np.ndarray
        Porosity aspect ratio.

    Returns
    -------
    tuple
        k, mu, rhob : np.ndarray
        k - effective bulk modulus [Pa], mu - effective shear modulus [Pa], rhob - effective density [kg/m^3].
    """

    #   tol: DEM model calculation tolerance <= Set as a hardcoded value, not
    #   found to influence the results
    tol = 1e-6

    # Calculate effective mineral properties
    if np.any((f1 + f2) > 1.0):
        raise ValueError(f"{__file__}: fixed mineral fractions exceed 1.0")
    if np.any(np.logical_or(f1 < 0, f2 < 0)):
        raise ValueError(f"{__file__}: negative mineral fractions")

    f3 = prop_clay * (1.0 - f1 - f2)
    f4 = (1.0 - prop_clay) * (1.0 - f1 - f2)
    k_mat, mu_mat = std_functions.multi_voigt_reuss_hill(
        k1, mu1, f1, k2, mu2, f2, k3, mu3, f3, k4, mu4, f4
    )
    rho_mat = rho1 * f1 + rho2 * f2 + rho3 * f3 + rho4 * f4

    k_mat, mu_mat, rho_mat, k_fl, rho_fl, phi, asp = cast(
        list[npt.NDArray[np.float64]],
        dim_check_vector((k_mat, mu_mat, rho_mat, k_fl, rho_fl, phi, asp)),
    )
    k, mu, rhob = dem_model(
        k1=k_mat,
        mu1=mu_mat,
        rho1=rho_mat,
        k2=k_fl,
        mu2=np.zeros(len(k_fl)),
        rho2=rho_fl,
        frac2=phi,
        asp2=asp,
        tol=tol,
    )

    return k, mu, rhob



import warnings
from typing import Literal, TypedDict, cast

import numpy as np

from rock_physics_open.equinor_utilities import gen_utilities
from rock_physics_open.equinor_utilities.various_utilities.types import Array1D, Array2D

from .array_functions import array_inverse, array_matrix_mult
from .calc_c_eff import calc_c_eff_visco_vec
from .calc_isolated import calc_isolated_part_vec
from .calc_kd import calc_kd_vec
from .calc_kd_uuv import calc_kd_uuvv_vec
from .calc_pressure import calc_pressure_vec
from .calc_td import calc_td_vec
from .calc_x import calc_x_vec
from .iso_av import iso_av_vec
from .pressure_input import pressure_input_utility
from .velocity_vti_angles import velocity_vti_angles_vec


class CaseDict(TypedDict):
    iso: Literal[0, 1, 2]
    con: Literal[0, 1, 2]


class ShapeParamsDict(TypedDict):
    alpha_con: Array1D[np.float64]
    alpha_iso: Array1D[np.float64]
    v_con: Array1D[np.float64]
    v_iso: Array1D[np.float64]
    tau_con: Array1D[np.float64]
    tau_iso: Array1D[np.float64]


def t_matrix_porosity_vectorised(
    k_min: Array1D[np.float64],
    mu_min: Array1D[np.float64],
    rho_min: Array1D[np.float64],
    k_fl: Array1D[np.float64],
    rho_fl: Array1D[np.float64],
    phi: Array1D[np.float64],
    perm: Array1D[np.float64],
    visco: Array1D[np.float64],
    alpha: Array1D[np.float64],
    v: Array1D[np.float64],
    tau: Array1D[np.float64],
    frequency: float,
    angle: float,
    frac_inc_con: float,
    frac_inc_ani: float,
    pressure: Array1D[np.float64] | None = None,
) -> tuple[
    Array2D[np.float64],
    Array2D[np.float64],
    Array2D[np.float64],
    Array2D[np.float64],
]:
    """Vectorised version of T-Matrix, pure Python version - mainly intended for cases where it is wished to follow
    the entire process through and study intermediate results. The C++ implementation is significantly faster.

    Description of inputs:
    Mineral properties (effective properties, assumed mixed).
    Fluid properties (effective properties, assume mixed).

    Parameters
    ----------
    k_min : np.ndarray
        N length array, bulk modulus of matrix/mineral [Pa].
    mu_min : np.ndarray
        N length array, shear modulus of matrix/mineral [Pa].
    rho_min : np.ndarray
        N length array, density of matrix/mineral [kg/m^3].
    k_fl : np.ndarray
        N length array, bulk modulus of fluid [Pa].
    rho_fl : np.ndarray
        N length array, density of fluid [kg/m^3].
    phi : np.ndarray
        N length array, porosity [fraction].
    perm : np.ndarray
        Single float or N length array, permeability [mD].
    visco : np.ndarray
        Single float or N length array, fluid viscosity [cP].
    alpha : np.ndarray
        M length vector, aspect ratio for inclusion sets [ratio].
    v : np.ndarray
        M length vector, fraction of porosity belonging to each inclusion set [fraction].
    tau : np.ndarray
        M length vector, relaxation time constant [s].
    frequency : float
        Single float, measurement frequency (seismic, sonic, ultrasonic range) [Hz].
    angle : float
        Single float, angle of symmetry plane (0 = HTI, 90 = VTI medium).
    frac_inc_con : float or np.ndarray
        Single float or N length array, fraction of inclusions that are connected.
    frac_inc_ani : float or np.ndarray
        Single float or N length array, fraction of inclusions that are anisotropic.
    pressure : np.ndarray, optional
        L length array (normally 2), by default None.

    Returns
    -------
    tuple
        Of type (np.ndarray, np.ndarray, np.ndarray, np.ndarray). Vertical P-wave velocity [m/s], Vsv: Vertical polarity S-wave velocity [m/s],
        Vsh: Horizontal polarity S-wave velocity [m/s], rho_b: bulk density [kg/m^3].
    """
    log_length = len(phi)
    # Check that the inputs that should have the same length actually do
    (
        k_min,
        mu_min,
        rho_min,
        k_fl,
        rho_fl,
        phi,
        perm,
        visco,
    ) = cast(  # casting since dim_check_vector typing is incomplete
        list[Array1D[np.float64]],
        gen_utilities.dim_check_vector(
            (k_min, mu_min, rho_min, k_fl, rho_fl, phi, perm, visco)
        ),
    )

    # Conversion to SI units
    perm = perm * 0.986923e-15
    visco = visco * 1.0e-2

    # Alpha, v and tau should be of the same length
    (alpha, v, tau) = cast(  # casting since dim_check_vector typing is incomplete
        list[Array1D[np.float64]],
        gen_utilities.dim_check_vector((alpha, v, tau)),
    )
    shape_len = alpha.shape[0]

    # Shape parameters go into a dict, duplicate here, can be changed with pressure effect
    shape_params: ShapeParamsDict = {
        "alpha_con": alpha,
        "alpha_iso": alpha,
        "v_con": v * frac_inc_con,
        "v_iso": v * (1 - frac_inc_con),
        "tau_con": tau,
        "tau_iso": tau,
    }
    # Create case based on amount of anisotropic inclusions
    if frac_inc_ani == 0:
        iso = 0
        angle = 0
    elif frac_inc_ani == 1:
        iso = 2
    else:
        iso = 1

    # Create case based on amount of connected inclusions
    if frac_inc_con == 0:
        con = 0
    elif frac_inc_con == 1:
        # All connected
        con = 2
    else:
        # Mixed case
        con = 1

    case: CaseDict = {"iso": iso, "con": con}

    pressure_steps = 1
    delta_pres = np.array(0.0)
    if pressure is not None:
        try:
            # Pressure should be an object with __len__ attribute, cast to np.array
            pressure = np.array(pressure)
            pressure_steps = len(pressure)
        except TypeError:
            w_str = "pressure input must be a vector of length > 1 with increasing effective pressures"
            warnings.warn(w_str)
        else:
            delta_pres = np.diff(pressure)

    # Predefine output vectors
    vp = np.zeros((log_length, pressure_steps))
    vs_v = np.zeros((log_length, pressure_steps))
    vs_h = np.zeros((log_length, pressure_steps))
    rho_b_est = np.zeros((log_length, pressure_steps))

    # Matrix properties needed
    c0, s0, gd = pressure_input_utility(k_min, mu_min, log_length)

    i4 = np.tile(np.eye(6).reshape(1, 6, 6), (log_length, 1, 1))

    # Vectorise v and alpha
    v_con = phi.reshape(log_length, 1) * shape_params["v_con"].reshape(1, shape_len)
    alpha_con = np.ones(v_con.shape) * shape_params["alpha_con"].reshape(1, shape_len)

    v_iso = phi.reshape(log_length, 1) * shape_params["v_iso"].reshape(1, shape_len)
    alpha_iso = np.ones(v_iso.shape) * shape_params["alpha_iso"].reshape(1, shape_len)

    for i in range(pressure_steps):
        # Check if v(j) > alpha(j)for maximum porosity. If true, set v(j) = alpha(j)/2 to make sure
        # the numbers of inclusions in the system is not violating the
        # approximations for effective medium theories.
        phi_con = 0.0
        phi_iso = 0.0
        if case["con"] != 0:
            idx_v_con = v_con * phi.reshape(log_length, 1) > alpha_con
            if np.any(idx_v_con):
                v_con[idx_v_con] = alpha_con[idx_v_con] / 2
            phi_con = np.sum(v_con, axis=1)
        if case["con"] != 2:
            idx_v_iso = v_iso * phi.reshape(log_length, 1) > alpha_iso
            if np.any(idx_v_iso):
                v_iso[idx_v_iso] = alpha_iso[idx_v_iso] / 2
            phi_iso = np.sum(v_iso, axis=1)
        # May seem unnecessary, but V-vectors can change with changing pressure
        phi_out = phi_con + phi_iso

        # Creating reference matrix
        vs_min = np.sqrt(mu_min / rho_min)

        if case["con"] == 0:
            # All isolated
            # Isolated part: calculated c1 tensor (sum over all the isolated t-matrices and concentrations
            c1 = calc_isolated_part_vec(
                c0=c0,
                s_0=s0,
                kappa_f=k_fl,
                alpha=alpha_iso,
                v=v_iso,
                case_iso=case["iso"],
                frac_ani=frac_inc_ani,
            )
            c_eff = c0 + array_matrix_mult(
                c1, array_inverse(i4 + array_matrix_mult(gd, c1))
            )
            gamma = np.zeros((log_length, 1))
        else:
            kd = calc_kd_vec(
                c0=c0,
                i4=i4,
                s_0=s0,
                alpha=alpha_con,
            )
            kd_uuvv = calc_kd_uuvv_vec(kd)
            gamma: Array2D[np.float64] = (1 - k_fl / k_min).reshape(
                log_length, 1
            ) + k_fl.reshape(log_length, 1) * kd_uuvv
            if case["con"] == 1:
                # Mix
                # Isolated part: calculated c1 tensor (sum over all the isolated t-matrices and concentrations
                c1 = calc_isolated_part_vec(
                    c0=c0,
                    s_0=s0,
                    kappa_f=k_fl,
                    alpha=alpha_iso,
                    v=v_iso,
                    case_iso=case["iso"],
                    frac_ani=frac_inc_ani,
                )
            else:  # case['con'] == 2
                # All connected - only c1 differs from the most general case
                c1 = np.zeros((log_length, 6, 6))
            # Connected part Calculate dry properties:
            td = calc_td_vec(
                c0=c0,
                i4=i4,
                s_0=s0,
                kd=kd,
                alpha=alpha_con,
            )
            # iso averaging the isotropic porosity
            td_bar = iso_av_vec(td)
            # Calculate the fluid effect
            x = calc_x_vec(s0, td)
            # iso averaging the isotropic porosity
            x_bar = iso_av_vec(x)
            # Frequency dependent stiffness
            c_eff = calc_c_eff_visco_vec(
                vs=vs_min,
                k_r=perm,
                eta_f=visco,
                v=v_con,
                gamma=gamma,
                tau=tau,
                kd_uuvv=kd_uuvv,
                kappa=k_min,
                kappa_f=k_fl,
                c0=c0,
                s0=s0,
                c1=c1,
                td=td,
                td_bar=td_bar,
                x=x,
                x_bar=x_bar,
                gd=gd,
                frequency=frequency,
                frac_ani=frac_inc_ani,
            )
        # Effective density
        rho_b_est[:, i] = phi_out * rho_fl + (1 - phi_out) * rho_min
        vp[:, i], vs_v[:, i], vs_h[:, i] = velocity_vti_angles_vec(
            c_eff=c_eff,
            rho_eff=rho_b_est[:, i],
            angle=angle,
        )

        if i != pressure_steps - 1:
            alpha_con, v_con, alpha_iso, v_iso, tau, gamma = calc_pressure_vec(
                alpha_con=alpha_con,
                alpha_iso=alpha_iso,
                v_con=v_con,
                v_iso=v_iso,
                c0=c0,
                s_0=s0,
                gd=gd,
                d_p=delta_pres[i],
                tau=tau,
                gamma=gamma,
                k_fl=k_fl,
                ctrl=case["con"],
                frac_ani=frac_inc_ani,
            )

    return vp, vs_v, vs_h, rho_b_est



import inspect
from pathlib import Path
from typing import Any

import numpy as np
import numpy.typing as npt

INITIATE = False

SYSTEM_AND_DEBUG_FCN = [
    "pydev",
    "ipython-input",
    "interactiveshell",
    "async_helpers",
    "handle_snapshots",
    "run_code",
    "run_ast_nodes",
    "run_cell_async",
    "_pseudo_sync_runner",
    "_run_cell",
    "run_cell",
    "add_exec",
    "do_add_exec",
    "add_exec",
    "ipython_exec_code",
    "console_exec",
    "do_it",
    "process_internal_commands",
    "_do_wait_suspend",
    "do_wait_suspend",
    "compare_snapshots",
    "handle_snapshots",
    "wrapper",
    "run_tests",
    "<module>",
]


def get_snapshot_name(
    step: int = 1,
    include_filename: bool = True,
    include_function_name: bool = True,
    include_extension: bool = True,
    include_snapshot_dir: bool = True,
) -> str:
    """
    Parameters
    ----------
    step: number of steps in the trace to collect information from
    include_snapshot_dir: absolute directory name included in snapshot name
    include_filename: whether to include filename in snapshot name
    include_function_name: whether to include function name in snapshot name
    include_extension: whether to include extension in snapshot name

    Returns
    -------
    name of snapshot file
    """
    trace = inspect.stack()
    for frame in trace[step:]:
        if not any(keyword in frame.function for keyword in SYSTEM_AND_DEBUG_FCN):
            break
    else:
        frame = trace[step]

    dir_name = Path(frame.filename).parents[1] / "data" / "snapshots"
    file_name = Path(frame.filename).stem if include_filename else ""
    function_name = frame.function if include_function_name else ""
    extension = ".npz" if include_extension else ""
    parts = [part for part in [file_name, function_name] if part]
    base_name = "_".join(parts) + extension
    return str(dir_name / base_name) if include_snapshot_dir else base_name


def store_snapshot(snapshot_name: str, *args: np.ndarray) -> bool:
    """
    Examples
    --------
    In case there are multiple arrays to store:
    store_snapshot(snapshot_name='snap_to_store.npz', *args)

    Important: If there is only one array to store:
    store_snapshot(snapshot_name='snap_to_store.npz', args)
    """
    try:
        np.savez(snapshot_name, *args)
    except IOError as e:
        raise IOError(f"Could not store snapshot {snapshot_name}: {e}")
    return True


def read_snapshot(snapshot_name: str) -> tuple[npt.NDArray[Any], ...]:
    try:
        with np.load(snapshot_name) as stored_npz:
            return tuple(stored_npz[arr_name] for arr_name in stored_npz.files)
    except IOError as e:
        raise ValueError(f"unable to load snapshot {snapshot_name}: {e}")



import numpy as np

from rock_physics_open.equinor_utilities.various_utilities.types import Array2D, Array4D


def calc_kd_uuvv_vec(kd: Array4D[np.float64]) -> Array2D[np.float64]:
    """Returns the sum of dry k_uuvv.

    Parameters
    ----------
    kd : np.ndarray
        The dry K-tensor (n, 6,6,(numbers of inclusions)) matrix.

    Returns
    -------
    np.ndarray
        Summed elements.

    """
    return np.sum(kd[:, :3, :3, :], axis=(1, 2))



from typing import Any, cast

import numpy as np

from rock_physics_open.equinor_utilities import gen_utilities
from rock_physics_open.equinor_utilities.optimisation_utilities import (
    gen_opt_routine,
    opt_param_to_ascii,
    save_opt_params,
)
from rock_physics_open.equinor_utilities.various_utilities.types import Array1D

from .curvefit_t_matrix_min import curve_fit_2_inclusion_sets

# Trade-off between calcite, dolomite and quartz, vs is weighted by this in order to make it count as much as vp
# in the optimisation
DEF_VP_VS_RATIO = 1.8


def t_matrix_optimisation_petec(
    k_min: Array1D[np.float64],
    mu_min: Array1D[np.float64],
    rho_min: Array1D[np.float64],
    k_fl: Array1D[np.float64],
    rho_fl: Array1D[np.float64],
    por: Array1D[np.float64],
    vp: Array1D[np.float64],
    vs: Array1D[np.float64],
    rhob: Array1D[np.float64],
    angle: float = 0.0,
    k_r: float = 50.0,
    eta_f: float = 1.0,
    tau: float = 1.0e-7,
    freq: float = 1.0e3,
    file_out_str: str = "opt_params_min.pkl",
    display_results: bool = False,
    well_name: str = "Unknown well",
    **opt_kwargs: Any,
) -> tuple[
    Array1D[np.float64],
    Array1D[np.float64],
    Array1D[np.float64],
    Array1D[np.float64],
    Array1D[np.float64],
    Array1D[np.float64],
    Array1D[np.float64],
    Array1D[np.float64],
]:
    """T-Matrix optimisation adapted to a case with detailed information available, such as in a development or production
    setting. Mineral and fluid composition should be known on a sample basis. Inclusion parameters are regarded as
    unknown and they are optimised for.

    Parameters
    ----------
    k_min :
        Effective mineral bulk modulus [Pa].
    mu_min :
        Effective mineral shear modulus [Pa].
    rho_min :
        Effective mineral bulk density [kg/m^3].
    k_fl :
        Effective fluid bulk modulus [Pa].
    rho_fl :
        Effective fluid density [kg/m^3].
    por :
        Inclusion porosity [ratio].
    vp :
        Compressional velocity log [m/s].
    vs :
        Shear velocity log [m/s].
    rhob :
        Bulk density log [kg/m^3].
    angle : float
        Angle of symmetry plane [degrees]
    k_r :
        Permeability [mD].
    eta_f :
        Fluid viscosity [cP].
    tau :
        Relaxation time constant [s].
    freq :
        Signal frequency [Hz].
    file_out_str :
        Output file name (string) to store optimal parameters (pickle format).
    display_results :
        Display optimal parameters in a window after run.
    well_name:
        Name of well to be displayed in info box title.
    opt_kwargs:
        Additional keywords to be passed to optimisation function

    Returns
    -------
    tuple
        vp_mod, vs_mod, rho_mod, ai_mod, vpvs_mod - modelled logs, vp_res, vs_res, rho_res - residual logs.
    """
    # 1. Preparation that is independent of search for minimum possible aspect ratio for second inclusion set

    # Optimisation function for selected parameters, with effective mineral properties known and 2 inclusion sets
    opt_fun = curve_fit_2_inclusion_sets
    rhob_mod = rho_min * (1 - por) + rho_fl * por
    rhob_res = rhob - rhob_mod
    # PETEC adapted inputs: include fluid data and other params in x_data
    por, angle_, k_r_, eta_f_, tau_, freq_, def_vp_vs_ratio = cast(
        list[Array1D[np.float64]],
        gen_utilities.dim_check_vector(
            (
                por,
                angle,
                k_r,
                eta_f,
                tau,
                freq,
                DEF_VP_VS_RATIO,
            )
        ),
    )
    x_data = np.stack(
        (
            por,
            k_min,
            mu_min,
            rho_min,
            k_fl,
            rho_fl,
            angle_,
            k_r_,
            eta_f_,
            tau_,
            freq_,
            def_vp_vs_ratio,
        ),
        axis=1,
    )
    # Set weight to vs to give vp and vs similar influence on optimisation
    y_data = np.stack([vp, vs * DEF_VP_VS_RATIO], axis=1)

    # 2. Search for minimum aspect ratio of inclusion set no. 2, given that inclusion set no. 1 will at least represent
    # 50% of inclusions. Minimum aspect ratio is linked to the porosity, and the most conservative estimate is to use
    # the maximum value. This is likely to deteriorate the optimisation results, so a search for a more appropriate
    # value that still produces valid results is made
    percentiles = [50, 75, 80, 85, 90, 95, 99, 100]
    valid_result = False
    vel_mod = None
    vel_res = None
    opt_params = None
    while not valid_result and percentiles:
        try:
            # Make sure that parameters are not in conflict with T Matrix assumptions
            min_v1 = 0.5
            max_por = np.percentile(por, percentiles[0])
            min_a2 = (1.0 - min_v1) * max_por
            # Test with all parameters in the range 0.0 - 1.0 for best optimiser performance
            # Params:               f_ani f_con a1   a2    v1
            lower_bound = np.array([0.0, 0.0, 0.5, min_a2, min_v1], dtype=float)
            upper_bound = np.array([1.0, 1.0, 1.0, 0.30, 1.0], dtype=float)
            x0 = (upper_bound + lower_bound) / 2.0
            # Optimisation step without fluid substitution
            vel_mod, vel_res, opt_params = gen_opt_routine(
                opt_function=opt_fun,
                x_data_orig=x_data,
                y_data=y_data,
                x_init=x0,
                low_bound=lower_bound,
                high_bound=upper_bound,
                **opt_kwargs,
            )
            valid_result = True
        except ValueError:
            _ = percentiles.pop(0)
            valid_result = False

    if not valid_result:
        raise ValueError(
            f"{__file__}: unable to find stable value for T Matrix optimisation, second inclusion"
        )
    if vel_mod is None or vel_res is None or opt_params is None:
        raise ValueError(
            f"{__file__}: expected optimisation routine to return values, but got None"
        )

    # Reshape outputs and remove weight from vs
    vp_mod, vs_mod = [arr.flatten() for arr in np.split(vel_mod, 2, axis=1)]
    ai_mod = vp_mod * rhob_mod
    vp_res, vs_res = [arr.flatten() for arr in np.split(vel_res, 2, axis=1)]
    vs_mod = vs_mod / DEF_VP_VS_RATIO
    vs_res = vs_res / DEF_VP_VS_RATIO
    vpvs_mod = vp_mod / vs_mod
    # Save the optimal parameters
    save_opt_params(
        opt_type="min",
        opt_params=opt_params,
        file_name=file_out_str,
        well_name=well_name,
    )
    if display_results:
        opt_param_to_ascii(in_file=file_out_str, well_name=well_name)

    return vp_mod, vs_mod, rhob_mod, ai_mod, vpvs_mod, vp_res, vs_res, rhob_res



from typing import cast

import numpy as np
import numpy.typing as npt

from rock_physics_open.equinor_utilities import std_functions
from rock_physics_open.equinor_utilities.gen_utilities import (
    dim_check_vector,
    filter_input_log,
    filter_output,
)

from .constant_cement_models import constant_cement_model_dry
from .friable_models import CoordinateNumberFunction, friable_model_dry

FRAC_CEM_UP = 0.1
P_EFF_LOW = 20.0e6


def constant_cement_model_pcm(
    kmin: npt.NDArray[np.float64],
    mymin: npt.NDArray[np.float64],
    kcem: npt.NDArray[np.float64],
    mycem: npt.NDArray[np.float64],
    kzero: npt.NDArray[np.float64],
    myzero: npt.NDArray[np.float64],
    phi: npt.NDArray[np.float64] | float,
    cem_frac: npt.NDArray[np.float64] | float,
    phic: npt.NDArray[np.float64] | float,
    n: float,
    red_shear: float,
):
    """
    kdry, mydry = contantcementmodel_pcm(kmin, mymin, kcem, mycem, kzero, myzero, phi, cem_frac, phic, n, red_shear)
    """

    #   Contact cement model (Dvorkin-Nur)for given cem_frac
    kcc, mycc = std_functions.dvorkin_contact_cement(
        frac_cem=cem_frac,
        por0_sst=phic,
        mu0_sst=mymin,
        k0_sst=kmin,
        mu0_cem=mycem,
        k0_cem=kcem,
        vs_red=red_shear,
        c=n,
    )

    #   Fraction of zero-porosity end member
    f1 = 1 - phi / (phic - cem_frac)

    #   Interpolating using Hashin -Shtrikman lower bound = Constant cement model.
    #   Same mineral point as upper and lower bound in patchy cement model
    kdry, mydry = std_functions.hashin_shtrikman_walpole(
        k1=kzero,
        mu1=myzero,
        k2=kcc,
        mu2=mycc,
        f1=f1,
        bound="lower",
    )

    return kdry, mydry


def patchy_cement_model_weight(
    k_min: npt.NDArray[np.float64],
    mu_min: npt.NDArray[np.float64],
    rho_min: npt.NDArray[np.float64],
    k_cem: npt.NDArray[np.float64],
    mu_cem: npt.NDArray[np.float64],
    rho_cem: npt.NDArray[np.float64],
    k_fl: npt.NDArray[np.float64],
    rho_fl: npt.NDArray[np.float64],
    phi: npt.NDArray[np.float64],
    p_eff: npt.NDArray[np.float64],
    frac_cem: npt.NDArray[np.float64] | float,
    phi_c: float,
    coord_num_func: CoordinateNumberFunction,
    n: float,
    shear_red: npt.NDArray[np.float64] | float,
    weight_k: npt.NDArray[np.float64] | float,
    weight_mu: npt.NDArray[np.float64] | float,
) -> tuple[
    npt.NDArray[np.float64],
    npt.NDArray[np.float64],
    npt.NDArray[np.float64],
    npt.NDArray[np.float64],
    npt.NDArray[np.float64],
]:
    """
    Patchy cement model for sands that are a combination of friable model and constant cement model. No fluid or
    pressure substitution. Input variables for weight of K and Mu determine the model's position between upper and
    lower bound.

    Parameters
    ----------
    k_min : np.ndarray
        Mineral bulk modulus [Pa].
    mu_min : np.ndarray
        Mineral shear modulus [Pa].
    rho_min : np.ndarray
        Mineral bulk density [kg/m^3].
    k_cem : np.ndarray
        Sandstone cement bulk modulus [Pa].
    mu_cem : np.ndarray
        Sandstone cement shear modulus [Pa].
    rho_cem : np.ndarray
        Cement bulk density [kg/m^3].
    k_fl : np.ndarray
        Fluid bulk modulus [Pa].
    rho_fl : np.ndarray
        Fluid bulk density [kg/m^3].
    phi : np.ndarray
        Total porosity [fraction].
    p_eff : np.ndarray
        Effective pressure [Pa].
    frac_cem : float
        Upper bound cement volume fraction [fraction].
    phi_c : float
        Critical porosity [fraction].
    coord_num_func : str
        Indication if coordination number should be calculated from porosity or kept constant.
    n : float
        Coordination number [unitless].
    shear_red : float
        Shear reduction factor for sandstone [fraction].
    weight_k : float
        Weight between friable and cemented model for bulk modulus.
    weight_mu : float
        Weight between friable and cemented model for shear modulus.

    Returns
    -------
    tuple
        k, mu, rhob, vp, vs : np.ndarray.
        vp :Saturated P-velocity [m/s] after fluid and pressure substitution,
        vs : Saturated S-velocity [m/s] after fluid and pressure substitution,
        rhob : Saturated density [kg/m3] after fluid and pressure substitution,
        k : Saturated rock bulk modulus [Pa],
        mu : Shear modulus [Pa].
    """

    k_zero, mu_zero = std_functions.hashin_shtrikman_walpole(
        k1=k_cem,
        mu1=mu_cem,
        k2=k_min,
        mu2=mu_min,
        f1=frac_cem,
        bound="lower",
    )

    # In this implementation of the patchy cement model the given cement fraction for the constant cement model defines
    # the upper bound, and the effective pressure for the friable model defines the lower bound

    k_fri, mu_fri = friable_model_dry(
        k_min=k_zero,
        mu_min=mu_zero,
        phi=phi,
        p_eff=p_eff,
        phi_c=phi_c,
        coord_num_func=coord_num_func,
        n=n,
        shear_red=shear_red,
    )

    k_up, mu_up = constant_cement_model_dry(
        k_min=k_min,
        mu_min=mu_min,
        k_cem=k_cem,
        mu_cem=mu_cem,
        phi=phi,
        frac_cem=frac_cem,
        phi_c=phi_c,
        n=n,
        shear_red=shear_red,
        return_k_zero=False,
        extrapolate_to_max_phi=True,
    )

    k_dry = k_fri + weight_k * (k_up - k_fri)
    mu = mu_fri + weight_mu * (mu_up - mu_fri)

    k = std_functions.gassmann(
        k_dry,
        por=phi,
        k_fl=k_fl,
        k_min=k_zero,
    )

    weight_rho = 0.5 * (weight_k + weight_mu)
    rhob = (
        phi * rho_fl
        + (1 - phi - frac_cem * weight_rho) * rho_min
        + frac_cem * weight_rho * rho_cem
    )

    vp, vs, ai, vpvs = std_functions.velocity(k=k, mu=mu, rhob=rhob)

    return vp, vs, rhob, ai, vpvs


def patchy_cement_model_cem_frac(
    k_min: npt.NDArray[np.float64],
    mu_min: npt.NDArray[np.float64],
    rho_min: npt.NDArray[np.float64],
    k_cem: npt.NDArray[np.float64],
    mu_cem: npt.NDArray[np.float64],
    rho_cem: npt.NDArray[np.float64],
    k_fl: npt.NDArray[np.float64],
    rho_fl: npt.NDArray[np.float64],
    phi: npt.NDArray[np.float64],
    p_eff: npt.NDArray[np.float64],
    frac_cem: float,
    phi_c: float,
    coord_num_func: CoordinateNumberFunction,
    n: float,
    shear_red: float,
) -> tuple[
    npt.NDArray[np.float64],
    npt.NDArray[np.float64],
    npt.NDArray[np.float64],
    npt.NDArray[np.float64],
    npt.NDArray[np.float64],
]:
    """
    Patchy cement model for sands that are a combination of friable model and constant cement model. No fluid or
    pressure substitution. In this implementation of the patchy cement model the given cement fraction for the constant
    cement model defines the upper bound, and the effective pressure for the friable model defines the lower bound

    Parameters
    ----------
    k_min : np.ndarray
        Mineral bulk modulus [Pa].
    mu_min : np.ndarray
        Mineral shear modulus [Pa].
    rho_min : np.ndarray
        Mineral bulk density [kg/m^3].
    k_cem : np.ndarray
        Sandstone cement bulk modulus [Pa].
    mu_cem : np.ndarray
        Sandstone cement shear modulus [Pa].
    rho_cem : np.ndarray
        Cement bulk density [kg/m^3].
    k_fl : np.ndarray
        Fluid bulk modulus [Pa].
    rho_fl : np.ndarray
        Fluid bulk density [kg/m^3].
    phi : np.ndarray
        Total porosity [fraction].
    p_eff : np.ndarray
        Effective pressure [Pa].
    frac_cem : float
        Upper bound cement volume fraction [fraction].
    phi_c : float
        Critical porosity [fraction].
    coord_num_func : str
        Indication if coordination number should be calculated from porosity or kept constant, either "ConstVal" or
        "PoreBased" [default]
    n : float
        Coordination number [unitless].
    shear_red : float
        Shear reduction factor for sandstone [fraction].

    Returns
    -------
    tuple
        vp, vs, rhob, ai, vpvs : np.ndarray.
        vp :Saturated P-velocity [m/s] after fluid and pressure substitution,
        vs : Saturated S-velocity [m/s] after fluid and pressure substitution,
        rhob : Saturated density [kg/m3] after fluid and pressure substitution,
        ai : Saturated rock acoustic impedance [kg/m3 * m/s] after fluid and pressure substitution,
        vpvs : Saturated rock velocity ratio [ratio].
    """

    k_dry, mu, _ = patchy_cement_model_dry(
        k_min=k_min,
        mu_min=mu_min,
        rho_min=rho_min,
        k_cem=k_cem,
        mu_cem=mu_cem,
        rho_cem=rho_cem,
        phi=phi,
        p_eff=p_eff,
        frac_cem=frac_cem,
        phi_c=phi_c,
        coord_num_func=coord_num_func,
        n=n,
        shear_red=shear_red,
    )
    k_zero, _ = std_functions.hashin_shtrikman_walpole(
        k1=k_cem,
        mu1=mu_cem,
        k2=k_min,
        mu2=mu_min,
        f1=FRAC_CEM_UP,
        bound="lower",
    )

    k = std_functions.gassmann(
        k_dry=k_dry,
        por=phi,
        k_fl=k_fl,
        k_min=k_zero,
    )

    rhob = phi * rho_fl + (1 - phi - frac_cem) * rho_min + frac_cem * rho_cem

    vp, vs, ai, vpvs = std_functions.velocity(k=k, mu=mu, rhob=rhob)

    return vp, vs, rhob, ai, vpvs


def patchy_cement_model_dry(
    k_min: npt.NDArray[np.float64],
    mu_min: npt.NDArray[np.float64],
    rho_min: npt.NDArray[np.float64],
    k_cem: npt.NDArray[np.float64],
    mu_cem: npt.NDArray[np.float64],
    rho_cem: npt.NDArray[np.float64],
    phi: npt.NDArray[np.float64],
    p_eff: npt.NDArray[np.float64],
    frac_cem: float,
    phi_c: float,
    coord_num_func: CoordinateNumberFunction,
    n: float,
    shear_red: float,
) -> tuple[npt.NDArray[np.float64], npt.NDArray[np.float64], npt.NDArray[np.float64]]:
    """
    Patchy cement model for sands that are a combination of friable model and constant cement model. No fluid or
    pressure substitution. In this implementation of the patchy cement model the given cement fraction for the constant
    cement model defines the upper bound, and the effective pressure for the friable model defines the lower bound

    Parameters
    ----------
    k_min : np.ndarray
        Mineral bulk modulus [Pa].
    mu_min : np.ndarray
        Mineral shear modulus [Pa].
    rho_min : np.ndarray
        Mineral bulk density [kg/m^3].
    k_cem : np.ndarray
        Sandstone cement bulk modulus [Pa].
    mu_cem : np.ndarray
        Sandstone cement shear modulus [Pa].
    rho_cem : np.ndarray
        Cement bulk density [kg/m^3].
    phi : np.ndarray
        Total porosity [fraction].
    p_eff : np.ndarray
        Effective pressure [Pa].
    frac_cem : float
        Upper bound cement volume fraction [fraction].
    phi_c : float
        Critical porosity [fraction].
    coord_num_func : str
        Indication if coordination number should be calculated from porosity or kept constant, either "ConstVal" or
        "PoreBased" [default]
    n : float
        Coordination number [unitless].
    shear_red : float
        Shear reduction factor for sandstone [fraction].

    Returns
    -------
    tuple
        k:dry, mu, rho_dry : np.ndarray
        k_dry: dry rock bulk modulus [Pa],
        mu : dry rock shear modulus [Pa],
        rho_dry : dry rock density [kg/m3].,
    """
    # There are cases which suffer from a lack of consistency check at this stage,
    # add dim_check_vector and filter input/output
    phi, k_min, mu_min, rho_min, k_cem, mu_cem, rho_cem, p_eff = cast(
        list[npt.NDArray[np.float64]],
        dim_check_vector((phi, k_min, mu_min, rho_min, k_cem, mu_cem, rho_cem, p_eff)),
    )
    (idx, (phi, k_min, mu_min, rho_min, k_cem, mu_cem, rho_cem, p_eff)) = cast(
        tuple[npt.NDArray[np.bool_], list[npt.NDArray[np.float64]]],
        filter_input_log((phi, k_min, mu_min, rho_min, k_cem, mu_cem, rho_cem, p_eff)),
    )

    k_zero, mu_zero = std_functions.hashin_shtrikman_walpole(
        k1=k_cem,
        mu1=mu_cem,
        k2=k_min,
        mu2=mu_min,
        f1=FRAC_CEM_UP,
        bound="lower",
    )

    k_low, mu_low = friable_model_dry(
        k_min=k_zero,
        mu_min=mu_zero,
        phi=phi,
        p_eff=P_EFF_LOW * np.ones_like(phi),
        phi_c=phi_c,
        coord_num_func=coord_num_func,
        n=n,
        shear_red=shear_red,
    )

    k_fri, mu_fri = friable_model_dry(
        k_min=k_zero,
        mu_min=mu_zero,
        phi=phi,
        p_eff=p_eff,
        phi_c=phi_c,
        coord_num_func=coord_num_func,
        n=n,
        shear_red=shear_red,
    )

    k_up, mu_up = constant_cement_model_dry(
        k_min=k_min,
        mu_min=mu_min,
        k_cem=k_cem,
        mu_cem=mu_cem,
        phi=phi,
        frac_cem=FRAC_CEM_UP,
        phi_c=phi_c,
        n=n,
        shear_red=shear_red,
        extrapolate_to_max_phi=True,
        return_k_zero=False,
    )

    # Special case for the constant cement model that represents the mean of the data
    k_cc, mu_cc = constant_cement_model_pcm(
        kmin=k_min,
        mymin=mu_min,
        kcem=k_cem,
        mycem=mu_cem,
        kzero=k_zero,
        myzero=mu_zero,
        phi=phi,
        cem_frac=frac_cem,
        phic=phi_c,
        n=n,
        red_shear=shear_red,
    )

    idwk = k_up == k_low
    idwmu = mu_up == mu_low

    weight_k = np.ones(k_zero.shape)
    weight_mu = np.ones(mu_zero.shape)

    weight_k[~idwk] = (k_cc[~idwk] - k_low[~idwk]) / (k_up[~idwk] - k_low[~idwk])
    weight_mu[~idwmu] = (mu_cc[~idwmu] - mu_low[~idwmu]) / (
        mu_up[~idwmu] - mu_low[~idwmu]
    )

    weight_mu = np.clip(weight_mu, 0.0, 1.0)
    weight_k = np.clip(weight_k, 0.0, 1.0)

    k_dry = k_fri + weight_k * (k_up - k_fri)
    mu = mu_fri + weight_mu * (mu_up - mu_fri)

    rho_dry = (1 - phi - frac_cem) * rho_min + frac_cem * rho_cem

    k_dry, mu, rho_dry = cast(
        list[npt.NDArray[np.float64]], filter_output(idx, (k_dry, mu, rho_dry))
    )

    return k_dry, mu, rho_dry



import os

import numpy as np

from rock_physics_open.equinor_utilities.snapshot_test_utilities import (
    INITIATE,
    compare_snapshots,
    get_snapshot_name,
    read_snapshot,
    store_snapshot,
)
from rock_physics_open.fluid_models import oil_properties
from rock_physics_open.fluid_models.oil_model.live_oil_density import live_oil_density
from rock_physics_open.fluid_models.oil_model.live_oil_velocity import live_oil_velocity

temp = 100.0 * np.linspace(0.8, 1.2, 101)
pres = 30.0e6 * np.linspace(0.8, 1.2, 101)
rho_0 = 850.0 * np.ones(101)
gor = 120.0 * np.ones(101)
gr = 0.7 * np.ones(101)
gor_co2 = 50.0 * np.ones(101)
gr_co2 = 1.52 * np.ones(101)


def test_oil_prop():
    args = oil_properties(temp, pres, rho_0, gor, gr)

    if not os.path.isfile(get_snapshot_name()) or INITIATE:
        _ = store_snapshot(get_snapshot_name(), *args)
    else:
        assert compare_snapshots(args, read_snapshot(get_snapshot_name()))


def test_live_oil_density():
    args = live_oil_density(temp, pres, rho_0, gor, gr)

    if not os.path.isfile(get_snapshot_name()) or INITIATE:
        _ = store_snapshot(get_snapshot_name(), args)
    else:
        assert compare_snapshots(args, read_snapshot(get_snapshot_name()))


def test_live_oil_velocity():
    args = live_oil_velocity(temp, pres, rho_0, gor, gr)

    if not os.path.isfile(get_snapshot_name()) or INITIATE:
        _ = store_snapshot(get_snapshot_name(), args)
    else:
        assert compare_snapshots(args, read_snapshot(get_snapshot_name()))


def test_oil_properties_float():
    """
    Make sure that input object type is reflected in output type
    """
    for results in [
        oil_properties(temp[0], pres[0], rho_0[0], gor[0], gr[0]),
        live_oil_density(temp[0], pres[0], rho_0[0], gor[0], gr[0]),
        live_oil_velocity(temp[0], pres[0], rho_0[0], gor[0], gr[0]),
    ]:
        if hasattr(results, "__iter__"):
            assert all(isinstance(arg, float) for arg in results)
        else:
            assert isinstance(results, float)



import pickle
from typing import Any, Self, final

import numpy as np
from typing_extensions import override

from .base_pressure_model import BasePressureModel


@final
class PolynomialPressureModel(BasePressureModel):
    """
    Polynomial pressure sensitivity model for velocity prediction.

    Uses polynomial function: v = v0 * P(p_depl) / P(p_in_situ)
    where P(p) is a polynomial function with specified coefficients.

    Input format (n,3): [velocity, p_eff_in_situ, p_eff_depleted]
    """

    def __init__(
        self,
        weights: list[float],
        model_max_pressure: float | None = None,
        description: str = "",
    ):
        """
        Initialize polynomial pressure model.

        Parameters
        ----------
        weights : list[float]
            Polynomial coefficients [unitless]. First element is constant term,
            second is linear coefficient, etc.
        model_max_pressure : float | None
            Maximum pressure for predict_max method [Pa].
        description : str
            Model description.
        """
        super().__init__(model_max_pressure, description)
        self._weights = weights

    @property
    def weights(self) -> list[float]:
        """Polynomial coefficients."""
        return self._weights

    @override
    def validate_input(self, inp_arr: np.ndarray) -> np.ndarray:
        """
        Validate input for polynomial model.

        Parameters
        ----------
        inp_arr : np.ndarray
            Input array to validate.

        Returns
        -------
        np.ndarray
            Validated input array.

        Raises
        ------
        ValueError
            If input format is invalid.
        """
        if not isinstance(inp_arr, np.ndarray):  # pyright: ignore[reportUnnecessaryIsInstance] | Kept for backward compatibility
            raise ValueError("Input must be numpy ndarray.")  # pyright: ignore[reportUnreachable] | Kept for backward compatibility
        if inp_arr.ndim != 2 or inp_arr.shape[1] != 3:
            raise ValueError(
                "Input must be (n,3): [velocity, p_eff_in_situ, p_eff_depleted]"
            )
        return inp_arr

    @override
    def predict_abs(self, inp_arr: np.ndarray, case: str = "in_situ") -> np.ndarray:
        """
        Calculate absolute velocity for specified pressure case.

        Parameters
        ----------
        inp_arr : np.ndarray
            Validated input array (n,3).
        case : str
            Pressure case: "in_situ" or "depleted".

        Returns
        -------
        np.ndarray
            Velocity values [m/s].
        """
        arr = self.validate_input(inp_arr)

        # Validate weights are set
        if not self._weights:
            raise ValueError('Field "weights" is not set.')

        vel = arr[:, 0]
        p_in_situ = arr[:, 1]
        p_depleted = arr[:, 2]

        # Create polynomial from weights
        polynomial_expr = np.polynomial.Polynomial(self._weights)

        # Select pressure based on case
        p_eff = p_in_situ if case == "in_situ" else p_depleted

        # Calculate velocity using polynomial pressure correction
        return vel * polynomial_expr(p_eff) / polynomial_expr(p_in_situ)

    @override
    def todict(self) -> dict[str, Any]:
        """Convert model to dictionary."""
        return {
            "weights": self._weights,
            "model_max_pressure": self._model_max_pressure,
            "description": self._description,
        }

    @override
    @classmethod
    def load(cls, file: str | bytes) -> Self:
        """Load polynomial model from pickle file."""
        with open(file, "rb") as f_in:
            d = pickle.load(f_in)

        return cls(
            weights=d["weights"],
            model_max_pressure=d["model_max_pressure"],
            description=d["description"],
        )



from typing import Literal

import numpy as np

from rock_physics_open.equinor_utilities.various_utilities.types import (
    Array1D,
    Array2D,
    Array3D,
    Array4D,
)

from .array_functions import array_inverse, array_matrix_mult
from .calc_isolated import calc_isolated_part_vec
from .g_tensor import g_tensor_vec


def calc_kd_eff_vec(
    c0: Array3D[np.float64],
    s_0: Array3D[np.float64],
    k_fl: Array1D[np.float64],
    alpha_con: Array2D[np.float64],
    alpha_iso: Array2D[np.float64],
    v_con: Array2D[np.float64],
    v_iso: Array2D[np.float64],
    gd: Array3D[np.float64],
    ctrl: Literal[0, 1, 2],
    frac_ani: float,
) -> tuple[Array4D[np.float64] | None, Array4D[np.float64] | None]:
    """Returns the effective dry K-tensor (6x6x(numbers of inclusions) matrix.
    If there is no connected or no isolated pores, the function returns a NaN for
    the case which is not considered. E.g. if only isolated pores, the kd_eff_connected = NaN.

    Note: When isolated pores, the pores are considered as filled when
    calculating the dry effective K-tensor.

    Parameters
    ----------
    c0 : np.ndarray
        Stiffness tensor of the host material (nx6x6 matrix).
    s_0 : np.ndarray
        Inverse of the stiffness tensor.
    k_fl : np.ndarray
        Bulk modulus of the fluid (n length vector).
    alpha_con : np.ndarray
        Aspect ratio of connected inclusions.
    alpha_iso : np.ndarray
        Aspect ratio of isolated inclusions.
    v_con : np.ndarray
        Concentration of connected pores.
    v_iso : np.ndarray
        Concentration of isolated pores.
    gd : np.ndarray
        Correlation function (nx6x6 matrix).
    ctrl : int
        0 :only isolated pores, 1 :both isolated and connected pores, 2 :only connected pores.
    frac_ani : float
        Fraction of anisotropic inclusions.

    Returns
    -------
    tuple
         kd_eff_isolated, kd_eff_connected: (np.ndarray, np.ndarray).

    Notes
    -----
    Equations used can be found in:
    Agersborg (2007), phd thesis:
    https://bora.uib.no/handle/1956/2422

    09.03.2012
    Remy Agersborg
    email: remy@agersborg.com

    Translated to Python and vectorised by Harald Flesche, hfle@equinor.com 2020.
    """
    log_len = c0.shape[0]
    c1dry = np.zeros((log_len, 6, 6))
    c2dry = np.zeros((log_len, 6, 6))

    kd_eff_isolated = None
    kd_eff_connected = None

    c1_isolated = None
    if ctrl != 2:
        c1_isolated = calc_isolated_part_vec(
            c0=c0,
            s_0=s_0,
            kappa_f=k_fl,
            alpha=alpha_iso,
            v=v_iso,
            case_iso=ctrl,
            frac_ani=frac_ani,
        )
        c1dry = c1dry + c1_isolated
        if alpha_iso.ndim == 1 and alpha_iso.shape[0] != c0.shape[0]:
            alpha_iso = np.tile(alpha_iso.reshape(1, alpha_iso.shape[0]), (log_len, 1))
        c2dry = c2dry + array_matrix_mult(c1_isolated, gd, c1_isolated)
    if ctrl != 0:
        c1_connected = calc_isolated_part_vec(
            c0, s_0, np.zeros_like(k_fl), alpha_con, v_con, ctrl, frac_ani
        )
        c1dry = c1dry + c1_connected
        if alpha_con.ndim == 1 and alpha_con.shape[0] != c0.shape[0]:
            alpha_con = np.tile(alpha_con.reshape(1, alpha_con.shape[0]), (log_len, 1))
        c2dry = c2dry + array_matrix_mult(c1_connected, gd, c1_connected)
        if c1_isolated is not None:
            c2dry = (
                c2dry
                + array_matrix_mult(c1_connected, gd, c1_isolated)
                + array_matrix_mult(c1_isolated, gd, c1_connected)
            )

    i4 = np.tile(np.eye(6).reshape(1, 6, 6), (log_len, 1, 1))
    c_eff_dry = c0 + array_matrix_mult(
        c1dry, array_inverse(i4 + array_matrix_mult(array_inverse(c1dry), c2dry))
    )
    temp = array_matrix_mult(
        i4,
        array_inverse(i4 + array_matrix_mult(array_inverse(c1dry), c2dry)),
        array_inverse(c_eff_dry),
    )

    # if only connected or mixed connected and isolated
    if ctrl != 0:
        kd_eff_connected = np.zeros((log_len, 6, 6, alpha_con.shape[1]))
        for j in range(alpha_con.shape[1]):
            g = g_tensor_vec(
                c0=c0,
                s_0=s_0,
                alpha=alpha_con[:, j],
            )
            kd_eff_connected[:, :, :, j] = array_matrix_mult(
                array_inverse(i4 + array_matrix_mult(g, c0)), temp
            )

    if ctrl != 2:
        kd_eff_isolated = np.zeros((log_len, 6, 6, alpha_iso.shape[1]))
        for j in range(alpha_iso.shape[1]):
            g = g_tensor_vec(
                c0=c0,
                s_0=s_0,
                alpha=alpha_iso[:, j],
            )
            kd_eff_isolated[:, :, :, j] = array_matrix_mult(
                array_inverse(i4 + array_matrix_mult(g, c0)), temp
            )

    return kd_eff_isolated, kd_eff_connected



import numpy as np

from rock_physics_open.equinor_utilities.std_functions import dvorkin_contact_cement


class TestDvorkinNur:
    def test_dvorkin_nur(self):
        frac_cem = np.linspace(0.01, 0.1, 10)
        por0_sst = 0.4 * np.ones(10)
        mu0_sst = 44e9 * np.ones(10)
        k0_sst = 36.8e9 * np.ones(10)
        mu0_cem = 32e9 * np.ones(10)
        k0_cem = 71e9 * np.ones(10)
        vs_red = 0.25 * np.ones(10)
        c = 9.0
        k, mu = dvorkin_contact_cement(
            frac_cem, por0_sst, mu0_sst, k0_sst, mu0_cem, k0_cem, vs_red, c
        )
        k_expected = np.array(
            [
                2.8039220,
                3.9241764,
                4.7745324,
                5.4851787,
                6.1065427,
                6.6644869,
                7.1743746,
                7.6462158,
                8.0869637,
                8.5016807,
            ]
        )
        mu_expected = np.array(
            [
                2.2131204,
                3.0849503,
                3.745499,
                4.2966804,
                4.7779701,
                5.2096136,
                5.6036376,
                5.9678782,
                6.3077783,
                6.6273009,
            ]
        )
        np.testing.assert_almost_equal(k / 1.0e9, k_expected)
        np.testing.assert_almost_equal(mu / 1.0e9, mu_expected)



import numpy as np
import numpy.typing as npt

from rock_physics_open.equinor_utilities import std_functions


def hs_average(
    k1: npt.NDArray[np.float64],
    mu1: npt.NDArray[np.float64],
    rhob1: npt.NDArray[np.float64],
    k2: npt.NDArray[np.float64],
    mu2: npt.NDArray[np.float64],
    rhob2: npt.NDArray[np.float64],
    f: npt.NDArray[np.float64],
) -> tuple[
    npt.NDArray[np.float64],
    npt.NDArray[np.float64],
    npt.NDArray[np.float64],
    npt.NDArray[np.float64],
    npt.NDArray[np.float64],
    npt.NDArray[np.float64],
    npt.NDArray[np.float64],
]:
    """
    BMix of two phases by Hashin-Shtrikman model. Derived properties are also returned.

    Parameters
    ----------
    k1 : np.ndarray
        k1 array.
    mu1 : np.ndarray
        mu1 array.
    rhob1 : np.ndarray
        rhob1 array.
    k2 : np.ndarray
        k2 array.
    mu2 : np.ndarray
        mu2 array.
    rhob2 : np.ndarray
        rhob2 array.
    f : float or np.ndarray
        f value or array.

    Returns
    -------
    tuple
        vp, vs, rhob, ai, vp_vs, k, mu : np.ndarray
        vp: compressional wave velocity [m/s], vs: shear wave velocity [m/s], ai: acoustic impedance [m/s x kg/m^3],
        vp_vs: velocity ratio [ratio], k: bulk modulus [Pa], mu: shear modulus [Pa]

    """

    k, mu = std_functions.hashin_shtrikman_average(k1, mu1, k2, mu2, f)

    rhob = rhob1 * f + rhob2 * (1 - f)

    vp, vs, ai, vp_vs = std_functions.velocity(k, mu, rhob)

    return vp, vs, rhob, ai, vp_vs, k, mu



from typing import Any


def dict_value_to_float(
    input_dict: dict[str, Any],
) -> dict[str, float | list[float]]:
    """
    Convert dictionary strings to floating point numbers. Each value can have multiple floats.

    Parameters
    ----------
    input_dict : dict
        Input dictionary.

    Returns
    -------
    dict
        Output dictionary.
    """

    for item in input_dict:
        if isinstance(input_dict[item], float):
            pass
        else:
            try:
                ff = float(input_dict[item])
                input_dict[item] = ff
            except ValueError:  # if a list or tuple is hidden in the string
                try:
                    ll = eval(input_dict[item])
                    ff = [float(i) for i in ll]
                    input_dict[item] = ff
                except ValueError:
                    raise ValueError(
                        "dict_value_to_float: not possible to convert value to float"
                    )

    return input_dict



"""
Package for calculating the residual helmholtz energy for CO2. The module uses sympy to evaluate the functions defined
in Span & Wagner [2], as well as its derivatives.
"""

from typing import Any, Callable, Literal

import numpy as np
import numpy.typing as npt
import sympy as sp

from .coefficients import coeff_vars


def residual_helmholtz_energy(
    delta_: npt.NDArray[np.float64],
    tau_: npt.NDArray[np.float64],
    diff_delta: int,
    diff_tau: int,
) -> npt.NDArray[np.float64]:
    """
    Equation 6.1 from Span & Wagner [2]. Calculates the residual helmholtz energy of co2 or its derivatives. tau_ and
    delta_ must have be numpy arrays of shape (N, 1). This allows for vectorization.

    :param delta_: Reduced density. Unit-less. numpy.ndarray with shape (N, 1)
    :param tau_: Inverse reduced temperature. Unit-less. numpy.ndarray with shape (N, 1)
    :param diff_delta: Degree of derivation wrt. delta. Integer.
    :param diff_tau: Degree of derivation wrt. tau. Integer.

    :return: Helmholtz free energy. Unit-less. numpy.ndarray with shape (N,)
    """
    _s1 = np.sum(
        _LAMBDIFIED_EXPRESSIONS[(s1, diff_delta, diff_tau)](tau_, delta_), axis=-1
    )
    _s2 = np.sum(
        _LAMBDIFIED_EXPRESSIONS[(s2, diff_delta, diff_tau)](tau_, delta_), axis=-1
    )
    _s3 = np.sum(
        _LAMBDIFIED_EXPRESSIONS[(s3, diff_delta, diff_tau)](tau_, delta_), axis=-1
    )
    _s4 = np.sum(
        _LAMBDIFIED_EXPRESSIONS[(s4, diff_delta, diff_tau)](tau_, delta_), axis=-1
    )

    return _s1 + _s2 + _s3 + _s4


# Define coefficients symbols. These should correspond one-to-one with variable names in the coefficient module
coeff_symbols = (
    n1,
    t1,
    d1,
    n2,
    d2,
    t2,
    c2,
    n3,
    d3,
    t3,
    alpha3,
    epsilon3,
    beta3,
    gamma3,
    n4,
    b4,
    a4,
    beta4,
    A4,
    B4,
    C4,
    D4,
) = sp.symbols(
    "n1 t1 d1 n2 d2 t2 c2 n3 d3 t3 alpha3 epsilon3 beta3 gamma3 n4 b4 a4 beta4 A4 B4 C4 D4"
)
tau, delta = sp.symbols("tau delta", real=True)
i = sp.symbols("i", integer=True)
s1 = n1 * delta**d1 * tau**t1
s2 = n2 * delta**d2 * tau**t2 * sp.exp(-(delta**c2))
s3 = (
    n3
    * delta**d3
    * tau**t3
    * sp.exp(-alpha3 * (delta - epsilon3) ** 2 - beta3 * (tau - gamma3) ** 2)
)

theta_expr = (1 - tau) + A4 * ((delta - 1) ** 2) ** (1 / (2 * beta4))
bigdelta_expr = theta_expr**2 + B4 * ((delta - 1) ** 2) ** a4
bigphi_expr = sp.exp(-C4 * (delta - 1) ** 2 - D4 * (tau - 1) ** 2)
s4 = n4 * bigdelta_expr**b4 * delta * bigphi_expr


def _lambdify(
    expr: sp.Expr,
    diff_delta: Literal[0, 1, 2],
    diff_tau: Literal[0, 1, 2],
) -> Callable[..., Any]:
    diff = [delta] * diff_delta + [tau] * diff_tau
    if len(diff) > 0:
        expr = expr.diff(*diff)  # pyright: ignore[reportUnknownVariableType] | Sympy typing is incomplete
    expr = expr.powsimp()  # pyright: ignore[reportUnknownVariableType, reportUnknownMemberType] | Sympy typing is incomplete
    _sympy_lambda = sp.utilities.lambdify(  # pyright: ignore[reportUnknownVariableType] | Sympy typing is incomplete
        list(coeff_symbols) + [tau, delta],
        expr,
        modules=["numpy", {"DiracDelta": lambda x: x == 0}],  # pyright: ignore[reportUnknownLambdaType] | Sympy typing is incomplete
    )
    return lambda _tau, _delta: _sympy_lambda(*coeff_vars, _tau, _delta)  # pyright: ignore[reportUnknownVariableType, reportUnknownLambdaType] | Sympy typing is incomplete


_LAMBDIFIED_EXPRESSIONS: dict[tuple[Any, int, int], Callable[..., Any]] = {
    (e, dd, dt): _lambdify(expr=e, diff_delta=dd, diff_tau=dt)
    for e in (s1, s2, s3, s4)
    for dd in (0, 1, 2)
    for dt in (0, 1, 2)
    if dd + dt <= 2
}
